﻿using UnityEditor.IMGUI.Controls;
using UnityEngine;

namespace Playground.EditorInternal
{
    public class ConditionTreeElement : UnityTreeView.TreeElement
    {
        [SerializeField] private Condition _condition;
        [SerializeField] bool _fatal;
        public Condition Condition
        {
            get => _condition;
            internal set => _condition = value;
        }
        public bool Fatal
        {
            get => _fatal;
            internal set => _fatal = value;
        }


        //[SerializeField] SceneObject context;
        //[SerializeField] SceneObject relative;
        internal ConditionTreeElement() { }
        internal ConditionTreeElement(Condition condition)
        {
            _condition = condition;
        }
    }
    public class ConditionMultiColumnHeader : MultiColumnHeader
    {
        GUIStyle[] columnStyles;
        public ConditionMultiColumnHeader(MultiColumnHeaderState state) : base(state)
        {
            columnStyles = new GUIStyle[state.columns.Length];
            for (int i = 0; i < columnStyles.Length; i++)
            {
                var textAnchor =  (TextAnchor)(3 + state.columns[i].headerTextAlignment);
                columnStyles[i] = new GUIStyle("Label") { alignment = textAnchor, fontSize = 15, fontStyle = FontStyle.Bold };  
            }
        }

        protected override void ColumnHeaderGUI( MultiColumnHeaderState.Column column, Rect headerRect, int columnIndex )
        {
            //6,7,8  Upper
            //3,4,5, Middle
            //0,1,2, Lower
            GUI.Label(headerRect, column.headerContent, columnStyles[columnIndex]);
        }
    }


    internal class SceneObjectConditionTree : UnityTreeView.ModeledTreeView<ConditionTreeElement>
    {
        static Texture2D[] icons = new Texture2D[]
        {
            PlaygroundEditorResources.LoadIcon("checkmark"),
            PlaygroundEditorResources.LoadIcon("cross"),
        };

        public static class Styles {
            public static GUIStyle goalBackground       = new GUIStyle(DefaultStyles.backgroundOdd) { normal = new GUIStyleState(){ background = PlaygroundEditorResources.LoadBackground("goal")} };
            public static GUIStyle goalBackground_alt   = new GUIStyle(DefaultStyles.backgroundOdd) { normal = new GUIStyleState(){ background = PlaygroundEditorResources.LoadBackground("goal_alt")} };
            public static GUIStyle fatalBackground      = new GUIStyle(DefaultStyles.backgroundOdd) { normal = new GUIStyleState(){ background = PlaygroundEditorResources.LoadBackground("fatal")} };
            public static GUIStyle fatalBackground_alt  = new GUIStyle(DefaultStyles.backgroundOdd) { normal = new GUIStyleState(){ background = PlaygroundEditorResources.LoadBackground("fatal_alt")} };
        }

        [SerializeField] SceneObject _context;
        
        public SceneObjectConditionTree(
            TreeViewState state, 
            MultiColumnHeader multiColumnHeader, 
            UnityTreeView.TreeModel<ConditionTreeElement> model) 
        : base(state, multiColumnHeader, model)
        {
            rowHeight = 32;
            customFoldoutYOffset = 8;
            columnIndexForTreeFoldouts = 2; //name
            SetupDepthsFromParentsAndChildren(model.root);
            Reload();
        }

        protected override bool CanStartDrag( CanStartDragArgs args ) => false;

        protected override void RowGUI(RowGUIArgs args)
        {
            var item = (UnityTreeView.TreeViewItem<ConditionTreeElement>)args.item;

            var style = item.data.Fatal ?
                ((args.row % 2 == 0) ? Styles.fatalBackground : Styles.fatalBackground_alt) :
                ((args.row % 2 == 0) ? Styles.goalBackground : Styles.goalBackground_alt);

            GUI.Box(args.rowRect, GUIContent.none, style);
            TypeIconCell(args.GetCellRect(0), item.data);
            EvaluationCell(args.GetCellRect(1), item.data);
            ConditionNameCell(args.GetCellRect(2), ref args);
        }

        void TypeIconCell(Rect cellRect, ConditionTreeElement item)
        {
            //CenterRectUsingSingleLineHeight(ref cellRect);
            if (item.icon)
                GUI.DrawTexture(cellRect, item.icon, ScaleMode.ScaleToFit);

        }
        void EvaluationCell(Rect cellRect, ConditionTreeElement item)
        {
            CenterRectUsingSingleLineHeight(ref cellRect);
            int evaluationIconIndex = item.Condition.Evaluate() ? 0 : 1;
            GUI.DrawTexture(cellRect, icons[evaluationIconIndex], ScaleMode.ScaleToFit);
        }
        void ConditionNameCell(Rect cellRect, ref RowGUIArgs args)
        {
            args.rowRect = new Rect(cellRect) { y = cellRect.y + 8, height = cellRect.height - 8 };
            //cellRect.x += 16 + depthIndentWidth * args.item.depth;
            //GUI.Label(cellRect, args.label);
            base.RowGUI(args);
        }

        //void DoFoldoutButton( Rect foldoutRect, bool expanded, GUIStyle foldoutStyle )
        //{
        //    return GUI.Toggle(foldoutRect, expanded, GUIContent.none, foldoutStyle);
        //}

        public static MultiColumnHeaderState CreateDefaultMultiColumnHeaderState(float width)
        {
            var columns = new[]
            {
                new MultiColumnHeaderState.Column
                {
                    headerContent = new GUIContent(PlaygroundEditorResources.LoadIcon("Condition"), "Type of condition."),
                    contextMenuText = "Asset",
                    headerTextAlignment = TextAlignment.Center,
                    sortedAscending = true,
                    sortingArrowAlignment = TextAlignment.Right,
                    width = 32,
                    minWidth = 32,
                    maxWidth = 32,
                    autoResize = true,
                    allowToggleVisibility = true
                },
                new MultiColumnHeaderState.Column
                {
                    headerContent = new GUIContent(PlaygroundEditorResources.LoadIcon("Evaluation"), "Evaluation state."),
                    contextMenuText = "Type",
                    headerTextAlignment = TextAlignment.Center,
                    sortedAscending = true,
                    sortingArrowAlignment = TextAlignment.Right,
                    width = 32,
                    minWidth = 32,
                    maxWidth = 32,
                    autoResize = true,
                    allowToggleVisibility = true
                },
                new MultiColumnHeaderState.Column
                {
                    headerContent = new GUIContent("Condition"),
                    headerTextAlignment = TextAlignment.Left,
                    sortedAscending = true,
                    sortingArrowAlignment = TextAlignment.Center,
                    width = width-64,
                    minWidth = width-64,
                    maxWidth = width-64,
                    autoResize = true,
                    allowToggleVisibility = false,
                    
                },
            };
            return new MultiColumnHeaderState(columns);
        }
    }
}