﻿using UnityEngine;
using UnityEditor;
using UnityEditor.IMGUI.Controls;
using System;
using System.Collections.Generic;

namespace Playground.EditorInternal
{
    
    [CustomEditor(typeof(SceneObject), true)]
    public class SceneObjectEditor : Editor
    {
		[NonSerialized] bool _treeViewInitialized;
        [SerializeField] TreeViewState _conditionTreeViewState;
		[SerializeField] MultiColumnHeaderState _multiColumnHeaderState;

        SceneObjectConditionTree _conditionTreeView;

        new SceneObject target;

        SerializedProperty _showingConditionProp;
        private void OnEnable()
        {
            target = (SceneObject)base.target;
            _showingConditionProp = serializedObject.FindProperty("_editor_showConditions"); //TODO: Use EditorPrefs instead?
        }


        public override void OnInspectorGUI()
        {
            InitIfNeeded();
            EditorGUILayout.Space();
            
            var controlRect = GUILayoutUtility.GetRect(100, _conditionTreeView.totalHeight, GUILayout.ExpandWidth(true));
            Rect rect = new Rect(controlRect) { x = controlRect.x - 14, width = EditorGUIUtility.currentViewWidth };
            DoTreeView(rect);
            EditorGUI.BeginChangeCheck();
            _showingConditionProp.boolValue = GUILayout.Toggle(_showingConditionProp.boolValue, "Toggle Condition Visibility", "Button");
            if(EditorGUI.EndChangeCheck())
            {
                if (_showingConditionProp.boolValue)
                {
                    foreach (var c in target.GetComponents<Condition>())
                        c.hideFlags &= ~HideFlags.HideInInspector;
                }
                else
                {
                    foreach (var c in target.GetComponents<Condition>())
                        c.hideFlags |= HideFlags.HideInInspector;
                }
                serializedObject.ApplyModifiedPropertiesWithoutUndo();
            }

        }


        void OnSelectionChange()
        {

        }

        void DoTreeView(Rect rect)
        {
            _conditionTreeView.multiColumnHeader.GetColumn(2).width = EditorGUIUtility.currentViewWidth - 64;
            _conditionTreeView.OnGUI(rect);
        }

        void InitIfNeeded()
        {
            if (!_treeViewInitialized)
            {
                // Check if it already exists (deserialized from window layout file or scriptable object)
                if (_conditionTreeViewState == null)
                    _conditionTreeViewState = new TreeViewState();

                bool firstInit = _multiColumnHeaderState == null;
                var headerState = SceneObjectConditionTree.CreateDefaultMultiColumnHeaderState(EditorGUIUtility.currentViewWidth);
                if (MultiColumnHeaderState.CanOverwriteSerializedFields(_multiColumnHeaderState, headerState))
                    MultiColumnHeaderState.OverwriteSerializedFields(_multiColumnHeaderState, headerState);
                _multiColumnHeaderState = headerState;

                var multiColumnHeader = new ConditionMultiColumnHeader(headerState) { canSort = false };
                if (firstInit)
                    multiColumnHeader.ResizeToFit();

                var treeModel = new UnityTreeView.TreeModel<ConditionTreeElement>(GetData());

                _conditionTreeView = new SceneObjectConditionTree(_conditionTreeViewState, multiColumnHeader, treeModel);
                
                _treeViewInitialized = true;
            }
        }

        Texture2D IconFromCondition(Condition condition)
        {
            return PlaygroundEditorResources.LoadIcon($"Conditions/{condition.TypeName}") ?? PlaygroundEditorResources.LoadIcon("Missing");
        }

        IList<ConditionTreeElement> GetData()
        {
            void RecurseConditions(ref List<ConditionTreeElement> list, ConditionGroup groupCondition, bool fatal, int depth = 1)
            {
                foreach (var condition in groupCondition)
                {
                    list.Add(
                        new ConditionTreeElement(condition) {
                            icon = IconFromCondition(condition),
                            displayName = condition.Name,
                            id = list.Count,
                            depth = depth,
                            Fatal = fatal
                        }
                    );
                    if (condition is ConditionGroup conditionGroup)
                    {
                        RecurseConditions(ref list, conditionGroup, fatal, depth+1);
                    }
                }
            }


            var conditions = new List<ConditionTreeElement>() {
                new ConditionTreeElement { id = 0, depth = -1, displayName = "Root" }
            };
            
            if (target)
            {
                conditions.Add(
                    new ConditionTreeElement(target.GoalConditions) {
                        icon = IconFromCondition(target.GoalConditions),
                        displayName = "Goal Conditions",
                        id = conditions.Count,
                        depth = 0,
                        Fatal = false
                    }
                );
                RecurseConditions(ref conditions, target.GoalConditions, false);

                conditions.Add(
                    new ConditionTreeElement(target.TerminatingConditions) {
                        icon = IconFromCondition(target.TerminatingConditions),
                        displayName = "Terminating Conditions",
                        id = conditions.Count,
                        depth = 0,
                        Fatal = true
                    }
                );
                RecurseConditions(ref conditions, target.TerminatingConditions, true);
            }
            return conditions;
        }

    }
}
//namespace Playground.EditorInternal
//{
//    using Object = UnityEngine.Object;

//    [CustomEditor(typeof(SceneObject), true)]
//    public class SceneObjectEditor : Editor
//    {
//        VisualElement _rootElement;
//        VisualTreeAsset _visualTree;

//        private void OnEnable()
//        {
//            _rootElement = new VisualElement();

//            var editorAssemblyDefinition = UnityEditor.Compilation.CompilationPipeline.GetAssemblyDefinitionFilePathFromAssemblyName("Playground.Editor");
//            var editorAssemblyRootPath = editorAssemblyDefinition.Remove(editorAssemblyDefinition.LastIndexOf('/'));
//            Debug.Log(editorAssemblyRootPath);
//            //TODO make a static class for Utility access of different paths like this:
//            _visualTree = AssetDatabase.LoadAssetAtPath<VisualTreeAsset>(
//                $"{editorAssemblyRootPath}/VisualTrees/SceneObjectEditor.uxml"
//            );

//            var stylesheet = AssetDatabase.LoadAssetAtPath<StyleSheet>(
//                $"{editorAssemblyRootPath}/VisualTrees/SceneObjectEditor.uss"
//            );
//            _rootElement.styleSheets.Add(stylesheet);
//        }

//        public override VisualElement CreateInspectorGUI()
//        {
//            var root = _rootElement;
//            root.Clear();

//            _visualTree.CloneTree(root);
//            return root;
//        }

//        //[SerializeField]
//        //private ConditionGroupEditor _goal, _terminating;

//        //private void Awake()
//        //{
//        //    _goal = new ConditionGroupEditor(serializedObject.FindProperty("_goalConditions"));
//        //    _terminating = new ConditionGroupEditor(serializedObject.FindProperty("_terminatingConditions"));
//        //}

//        //public override void OnInspectorGUI()
//        //{
//        //    _goal.OnGUI();
//        //    EditorGUILayout.Space();
//        //    _terminating.OnGUI();
//        //}

//    }


//    //[Serializable]
//    //public class ConditionGroupEditor 
//    //{
//    //    private SerializedObject _target;
//    //    private SerializedProperty _conditions;

//    //    [SerializeField]
//    //    private bool _foldout;
//    //    private string _name;

//    //    public ConditionGroupEditor(SerializedProperty conditionGroup)
//    //    {
//    //        _target = new SerializedObject(conditionGroup.objectReferenceValue);
//    //        _conditions = _target.FindProperty("_conditions");
//    //        _name = conditionGroup.displayName;
//    //    }

//    //    public void OnGUI()
//    //    {
//    //        _foldout = EditorGUILayout.BeginFoldoutHeaderGroup(_foldout && _conditions.arraySize > 0, EditorGUIUtility.TrTextContent(_name));
//    //        if(_foldout)
//    //        {
//    //            for (int i = 0; i < _conditions.arraySize; i++)
//    //            {
//    //                EditorGUILayout.PropertyField(_conditions.GetArrayElementAtIndex(i));
//    //            }
//    //        }
//    //        EditorGUILayout.EndFoldoutHeaderGroup();

//    //    }
//    //}
//}