﻿using UnityEditor;
using UnityEngine;

namespace Playground.EditorInternal
{
    [CustomEditor(typeof(UI.WidgetCollection))]
    public class WidgetCollectionEditor : Editor
    {
        private UI.WidgetCollection _widgets;

        SerializedProperty keys, values;

        private const string    KEY_STR     = "_widgets._keys", 
                                VALUE_STR   = "_widgets._values";

        private void OnEnable()
        {
            _widgets = (UI.WidgetCollection)target;
            keys = serializedObject.FindProperty(KEY_STR);
            values = serializedObject.FindProperty(VALUE_STR);
        }

        public override void OnInspectorGUI()
        {
            if (keys.arraySize == 0)
                GUILayout.Box(
                    EditorGUIUtility.TrTextContent("No Entries In Dictionary", "There's nothing in this dictionary", "console.infoicon"),
                    GUILayout.ExpandWidth(true)
                );
            for (int i = 0; i < keys.arraySize; i++)
            {
                GUILayout.BeginHorizontal("Box");
                GUILayout.BeginVertical();
                EditorGUIUtility.labelWidth = 48;
                EditorGUI.BeginChangeCheck();
                EditorGUILayout.PropertyField(keys.GetArrayElementAtIndex(i), EditorGUIUtility.TrTempContent("Key"));
                EditorGUILayout.PropertyField(values.GetArrayElementAtIndex(i), EditorGUIUtility.TrTempContent("Value"));
                if (EditorGUI.EndChangeCheck())
                {
                    serializedObject.ApplyModifiedProperties();
                }
                GUILayout.EndVertical();
                if (GUILayout.Button(EditorGUIUtility.TrIconContent("Toolbar Minus"), GUILayout.Width(32), GUILayout.Height(32)))
                {
                    _widgets.Remove(keys.GetArrayElementAtIndex(i).stringValue);
                    var newState = new SerializedObject(target);
                    serializedObject.CopyFromSerializedProperty(newState.FindProperty(KEY_STR));
                    serializedObject.CopyFromSerializedProperty(newState.FindProperty(VALUE_STR));
                    serializedObject.ApplyModifiedProperties();
                }
                GUILayout.EndHorizontal();
            }
            
            DoDragAndDrop(GUILayoutUtility.GetRect(0, 80));
        }

        private GUIStyle HeaderText(int fontSize, Color color)
        {
            return new GUIStyle()
            {
                fontSize = fontSize,
                normal = new GUIStyleState() { textColor = color },
                alignment = TextAnchor.MiddleCenter
            };
        }
        private void DoDragAndDrop(in Rect pos)
        {
            Event e = Event.current;
            if (!pos.Contains(e.mousePosition))
            {
                //GUI.color = Color.gray;
                //Handles.DrawDottedLine(pos.min, new Vector3(pos.xMin, pos.yMax), 4);
                //Handles.DrawDottedLine(new Vector3(pos.xMin, pos.yMax), pos.max, pos.width / 50);
                //Handles.DrawDottedLine(pos.max, new Vector3(pos.xMax, pos.yMin), 4);
                //Handles.DrawDottedLine(new Vector3(pos.xMax, pos.yMin), pos.min, pos.width / 50);
                GUI.Box(pos, GUIContent.none, "ShurikenEffectBg");
                //GUI.color = Color.white;
                GUI.Label(
                    pos,
                    EditorGUIUtility.TrTextContent("Add Widget ...", tooltip: "Drag Widget prefabs here to add to dictionary"),
                    HeaderText(20, Color.white)
                );
                return;
            }
            GUI.color = Color.green;
            GUI.Box(pos, GUIContent.none, "ShurikenEffectBg");
            GUI.color = Color.white;
            GUI.Label(
                pos,
                EditorGUIUtility.TrTextContent("Add Widget ...", tooltip: "Drag Widget prefabs here to add to dictionary"),
                HeaderText(20, Color.white)
            );
            switch (e.type)
            {
                case EventType.DragUpdated:
                DragAndDrop.visualMode = DragAndDropVisualMode.Link;
                e.Use();
                return;
                case EventType.DragPerform:
                {
                    for (int i = 0; i < DragAndDrop.objectReferences.Length; i++)
                    {
                        UI.Widget widget;
                        var assetPath = PrefabUtility.GetPrefabAssetPathOfNearestInstanceRoot(DragAndDrop.objectReferences[i]);
                        if (assetPath.Length > 0)
                        {
                            widget = AssetDatabase.LoadAssetAtPath<GameObject>(assetPath)?.GetComponent<UI.Widget>();
                            if (widget)
                            {
                                Undo.RecordObject(target, $"Added Widget: {widget.GetType().FullName}");
                                _widgets.Add(widget.name, widget);
                                var newState = new SerializedObject(target);
                                serializedObject.CopyFromSerializedProperty(newState.FindProperty(KEY_STR));
                                serializedObject.CopyFromSerializedProperty(newState.FindProperty(VALUE_STR));
                                serializedObject.ApplyModifiedProperties();
                            }
                        }
                        else Debug.LogWarning($"Couldn't add <i>{ DragAndDrop.objectReferences[i].name }</i> to widget because it's not a widget prefab!");
                    }
                    e.Use();
                    return;
                }
                default: return;
            }
        }
    }
}
