﻿#if UNITY_EDITOR
using System.Collections.Generic;
using System.IO;
using UnityEditor;
using UnityEngine;
using UnityEngine.SceneManagement;

namespace Playground.EditorInternal
{
    public class SavesWindow : EditorWindow
    {
        private Vector2 scrollPos;
        private string currentSceneName = "Could not retrieve current scene";

        private bool saveNewProfileGroup;
        private bool loadProfileOnStart;
        private string savesPath;
        private string newProfileName = "";
        [SerializeField] private List<string> savedProfiles;
        [SerializeField] private string selectedProfile = "None";

        // Menu items to show window
        [MenuItem("Window/Playground/Saves")]
        [MenuItem("Tools/Playground/Saves")]

        public static void ShowWindow()
        {
            GetWindow(typeof(SavesWindow));
        }

        private void OnEnable()
        {
            titleContent = new GUIContent("Playground:Saves", "Window for loading saved environments with conditions set up");

            // Get current scene name
            Scene currentScene = SceneManager.GetActiveScene();
            currentSceneName = currentScene.name;

            selectedProfile = "None";

            LoadProfiles();
        }

        private void OnInspectorUpdate()
        {
            LoadProfiles();
        }

        private void OnGUI()
        {
            var serializedObject = new SerializedObject(this); // To serialize entries in the Window

            EditorGUILayout.BeginHorizontal(GUILayout.MaxWidth(100));
            EditorGUILayout.LabelField("Current scene: ", EditorStyles.boldLabel);
            EditorGUILayout.LabelField(currentSceneName, EditorStyles.label);
            EditorGUILayout.EndHorizontal();


            EditorGUILayout.LabelField("Profiles available:", EditorStyles.boldLabel);

            var savedProfileNames = serializedObject.FindProperty("savedProfiles");
            EditorGUILayout.BeginHorizontal(GUILayout.MaxWidth(800));
            scrollPos = EditorGUILayout.BeginScrollView(scrollPos, GUILayout.MaxWidth(500), GUILayout.Height(100));
            if (savedProfiles.Count == 0)
            {
                EditorGUILayout.LabelField("Could not find any profiles for this scene", EditorStyles.label);
            }
            else
            {
                for (int i = 0; i < savedProfiles.Count; i++)
                {
                    EditorGUILayout.BeginHorizontal();
                    // If this button is clicked, we set it as the current selected
                    if (GUILayout.Button(savedProfileNames.GetArrayElementAtIndex(i).stringValue))
                    {
                        selectedProfile = savedProfiles[i];
                        SaveSystem.ConditionDataManager.Instance.SelectedProfileName = selectedProfile;
                    }
                    if(GUILayout.Button("Delete", GUILayout.MaxWidth(100)))
                    {
                        SaveSystem.ConditionDataManager.Instance.DeleteProfile(savedProfiles[i]);
                        if (selectedProfile == savedProfiles[i])
                            selectedProfile = "None";
                    }
                    EditorGUILayout.EndHorizontal();
                }
            }
            EditorGUILayout.EndScrollView();

            if(GUILayout.Button("Open saves folder", GUILayout.MaxWidth(160), GUILayout.MaxHeight(100)))
            {
                EditorUtility.RevealInFinder(savesPath);
            }

            EditorGUILayout.EndHorizontal();

            EditorGUILayout.BeginHorizontal(GUILayout.MaxWidth(300));
            EditorGUILayout.LabelField("Current selected profile: ", EditorStyles.label);
            // TODO: Put ConditionDataManager.Instance.SelectedProfile instead of selectedProfile here?
            EditorGUILayout.LabelField(SaveSystem.ConditionDataManager.Instance.SelectedProfileName ?? selectedProfile, EditorStyles.label);
            EditorGUILayout.EndHorizontal();
            SaveSystem.ConditionDataManager.Instance.LoadProfileOnLoad = EditorGUILayout.Toggle("Load profile on start?", 
                                                                                                SaveSystem.ConditionDataManager.Instance.LoadProfileOnLoad);

            if(GUILayout.Button("Reload scene with selected profile", GUILayout.MaxWidth(200), GUILayout.MaxHeight(50)))
            {
                SaveSystem.ConditionDataManager.Instance.LoadProfileOnLoad = true;
                if (Application.isPlaying)
                {
                    SaveSystem.ConditionDataManager.Instance.ClearIDs(); // Hack to make clear ID's so they match on next load
                    SceneManager.LoadScene(currentSceneName);
                }
            }

            saveNewProfileGroup = EditorGUILayout.BeginToggleGroup("Save new profile?", saveNewProfileGroup);
            EditorGUILayout.BeginHorizontal(GUILayout.MaxWidth(600));
            newProfileName = EditorGUILayout.TextField("Profile name: ", newProfileName);
            if(GUILayout.Button("Save profile", GUILayout.MaxWidth(200)))
            {
                if(newProfileName != "")
                    SaveSystem.ConditionDataManager.Instance.SaveScene(newProfileName);
            }
            EditorGUILayout.EndToggleGroup();
            EditorGUILayout.EndHorizontal();

            //TODO: Have a bool that can be toggled that will choose to load or not load the current profile on start
            //This bool should be in an object in the environment ("ProfileLoader/Saver")

            serializedObject.ApplyModifiedProperties(); // Updating values that have changed
        }

        private void LoadProfiles()
        {
            // Get path for saved profiles
            savesPath = Path.Combine(Application.persistentDataPath, SaveSystem.ConditionDataManager.Instance.FolderName);
            savesPath = Path.Combine(savesPath, currentSceneName);

            string[] files = new string[0];
            try
            {
                files = Directory.GetFiles(savesPath);
            }
            catch (DirectoryNotFoundException e)
            {
                Debug.Log("Could not find any files for this scene at path: " + savesPath);
                Debug.Log(e.Message);
            }

            // Load each profile name into a list
            savedProfiles = new List<string>();
            if (files.Length > 0)
            {
                foreach (string file in files)
                {
                    if (file.EndsWith(".psbd")) // .PlaygroundSceneBinaryData
                    {
                        string[] split = file.Split("\\".ToCharArray());
                        string profileToAdd = split[split.Length - 1];                      // Split path to get only the entry in the directory
                        //profileToAdd = profileToAdd.Remove(0, currentSceneName.Length + 1); // Remove scene name
                        profileToAdd = profileToAdd.Remove(profileToAdd.Length - 5, 5);     // Remove .psbd
                        savedProfiles.Add(profileToAdd);
                    }
                }
            }
        }
    }
}
#endif