﻿using System;
using UnityEngine;

namespace Playground
{
    /// <summary>
    /// The base class for Conditions
    /// </summary>
    [Serializable]
    [HelpURL("https://justworksltd.gitlab.io/playground-docs/class_playground_1_1_condition.html")]
    public abstract class Condition : MonoBehaviour
    {
#if UNITY_EDITOR
        private void Reset() => hideFlags = HideFlags.HideInInspector;
#endif
        public uint Id;
        public virtual Type DataContainer => typeof(ConditionContainer); 
        /// <summary>
        /// The context SceneObject this condition is evaluated for
        /// </summary>
        [SerializeField, HideInInspector] private SceneObject _context;
        public SceneObject Context => _context;

        public virtual string Name => GetType().Name;
        public string TypeName => GetType().Name;

        /// <summary>
        /// Override InternalInitialize in Conditions to cache aside 
        /// references to components on the context SceneObject
        /// <para>This is called when the condition is first created, before </para>
        /// </summary>
        protected virtual void PreConfigure() { }

        protected virtual void PostConfigure() { }

        /// <summary>
        /// The entrypoint of initialization for all conditions
        /// <para>This is called when the condition is first created</para>
        /// </summary>
        /// <param name="ctx">The SceneObject this condition is placed on</param>
        /// <param name="newId">used by the loader to send persistent id. avoid using manually</param>
        public void Initialize(SceneObject ctx, uint newId = 0)
        {
            _context = ctx;

            if (newId == 0)
                Id = SaveSystem.ConditionDataManager.Instance.GetUniqueId();
            else
                Id = newId;

            PreConfigure();
        }
        /// <summary>
        /// The evaluationfunction to determine if the condition is satisfied
        /// </summary>
        /// <returns><c>true</c> if satisfied, <c>false</c> otherwise</returns>
        public abstract bool Evaluate();

        public virtual bool HasRelative => false;

        public virtual Vector3 RelativeDirection(in Vector3 worldSpaceDirection) => worldSpaceDirection;
        public virtual Vector3 RelativeVector(in Vector3 worldSpaceVector) => worldSpaceVector;
        public virtual Vector3 RelativePoint(in Vector3 worldSpacePoint) => worldSpacePoint;
    
        public virtual ConditionContainer GetDataContainer(ConditionValue conditionValue, bool inNestedGroupParam)
        {
            return new ConditionContainer
            {
                value = conditionValue,
                thisId = Id,
                contextId = _context.Id,
                inNestedGroup = inNestedGroupParam
            };
        }
        
    }

    [Serializable]
    public abstract class Condition<TRelative> : Condition, Internal.IHasRelative where TRelative : Component
    {
        public override Vector3 RelativeDirection(in Vector3 worldSpaceDirection)
        {
            return Relative.transform.InverseTransformDirection(worldSpaceDirection);
        }
        public override Vector3 RelativeVector(in Vector3 worldSpaceVector)
        {
            return Relative.transform.InverseTransformVector(worldSpaceVector);
        }
        public override Vector3 RelativePoint(in Vector3 worldSpacePoint)
        {
            return Relative.transform.InverseTransformPoint(worldSpacePoint);
        }
        /// <summary>
        /// The condition relative.
        /// <para>
        /// If a condition is dependent on a specific relation to a component, 
        /// it should inherit from a <see cref="Condition&#60;TRelative&#62;"/> with 
        /// <see cref="TRelative"/> set to that component
        /// </para>
        /// see <see cref="TouchCondition"/> for an example of usage.
        /// </summary>
        [SerializeField] private TRelative _relative;
        public TRelative Relative => _relative;

        public override bool HasRelative => _relative != null;

        Component Internal.IHasRelative.Relative => _relative;
        public Type RelativeType => typeof(TRelative);
        public void SetRelative(Component relativeComponent) => _relative = (TRelative)relativeComponent;

        public override ConditionContainer GetDataContainer(ConditionValue conditionValue, bool inNestedGroupParam)
        {
            return new ConditionContainer
            {
                value = conditionValue,
                thisId = Id,
                contextId = Context.Id,
                inNestedGroup = inNestedGroupParam,
                relativeId = ((_relative != null) ? _relative.GetComponent<SceneObject>().Id : 0)
            };
        }
    }
}
