﻿using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;
using Valve.VR;

namespace Playground
{
    /// <summary>Sinlgeton class that keeps the state of the scene and holds references to important objects.
    /// Modes: 
    /// <list type="bullet">
    /// <item>Annotation
    /// <description>The user annotates what conditions exist on spesific objects used in the demonstration.</description>
    /// </item>
    /// <item>Traing
    /// <description>The python learning process starts training on the generated reverse curriculum points.</description>
    /// </item>
    /// </list>
    /// </summary>
    public class GameManager : MonoBehaviour
    {
        static GameManager _instance;

        const float _buttonTimeout = 0.3f;

#pragma warning disable 0649
        [SerializeField] private bool _debug = false;
        [SerializeField] public bool usingVR = true;    // If the scene is not set up for VR(eg, missing player prefab), uncheck this. Unchecks on start if player can't be found.
        [SerializeField] public Transform player;       // Refernce to the player. If not using VR it should be set to a transform mocking the player.
        [SerializeField] private GameObject _tabletPrefab;
        [SerializeField] private GameObject _laserPointerPrefab;
        [SerializeField] private GameObject _actor;      // The actor in the scene (usually Franka robot) that should be disabled and enabled for training

        public UI.TransformFollow tabletMenuHandFollow;
        public UI.TransformFollow laserPointerHandFollow;
        public Valve.VR.InteractionSystem.Hand leftHand;// { get; private set; }
        public Valve.VR.InteractionSystem.Hand rightHand;// { get; private set; }
#pragma warning restore 0649

        public Canvas Canvas { get; private set; }

        public Internal.RaycastingInputModule rayInputModule { get; private set; }
        public UI.TabletMenu tabletMenu { get; private set; }
        public UI.LaserPointer laserPointer { get; private set; }

        private int _termCount;
        public int terminatingCount { get { return _termCount; } }
        private int _goalCount;
        public int goalCount { get { return _goalCount; } }
        public int sceneObjectsCount { get { return (_sceneObjects != null) ? _sceneObjects.Count : 0; } }
        public bool actorExist => (_actor != null);

        private List<GameObject> _annotationObjects;     // The objects in use when annotating conditions. Menus, laserpointer, ect.
        private List<GameObject> _trainingObjects;       // The objects in use when training the python learing process in the neo part. Robot hand, cameras.
        public List<SceneObject> _sceneObjects;          // The interactable objects the user picks up and demonstrates a task with.
        public IReadOnlyList<SceneObject> SceneObjects => _sceneObjects;
        private SteamVR_Action_Boolean _interactUIButton = SteamVR_Input.GetAction<SteamVR_Action_Boolean>("default", "MenuButton");
        private float _buttonTimeoutLeft;

        public static GameManager Instance
        {
            get                                         // On get gameManager instance, of instance not exist, make instance.
            {
                if (!_instance)
                    _instance = FindObjectOfType<GameManager>();
                if (!_instance)
                    _instance = new GameObject("GameManager", typeof(GameManager)).GetComponent<GameManager>();

                return _instance;
            }
        }

        /// <summary>
        /// Destroys itself if a GameManager already exist, then creates the relevant containers lastly runs Setup method.
        /// </summary>
        private void Awake()
        {
            if (Instance != this)           // Ensure that another instance of the gameManager cannot be forced by making a second GameManager object.
                Destroy(gameObject);

            DontDestroyOnLoad(gameObject); // TODO: This causes ConditionDataManager to not load profiles correct
            _annotationObjects = new List<GameObject>();
            _trainingObjects = new List<GameObject>();
            _sceneObjects = new List<SceneObject>();

            // TODO: This should probably be run in a delegate OnSceneLoaded in case there is no saved profile that should be loaded
            // as then this setup will not be recalled on scene reload
            Setup();
        }


        /// <summary>
        /// Runs teardown method to clear old references and empty the lists. Then finds all the refences again. 
        /// Objects it finds:
        /// <list type="bullet">
        /// <item>ConditionMenu controllers
        /// <description>These are prototyping menus used in VR to add conditions to a SceneObject with relative reference.</description>
        /// </item>
        ///<item>ConditionMenus
        /// <description>The corresponding menus the ConditionMenuControllers use.</description>
        ///</item>
        ///<item>Annotation objects
        /// <description>All GameObjects in the scene with the tag "Annotation". These are typically disabled in AI training.</description>
        ///</item>
        ///<item>SceneObjects
        /// <description>All GameObjects That can have conditions. This list is used to evaluate the state of the scene.</description>
        ///</item>
        ///<item>Actor
        /// <description>Actor GameObject in the scene.</description>
        ///</item>
        ///<item>Player
        /// <description>The transform of the root object in the VR player. If one can't be found, or if usingVR is set to false, a mock player will be created instead.</description>
        ///</item>
        ///<item>VR player's hands
        /// <description>References to the Hand script on the left and right hand in the VR player.</description>
        ///</item>
        ///<item>TabletMenu
        /// <description>The floating Menu the player uses to interact with the condition system.</description>
        ///</item>
        ///<item>LaserPointer
        /// <description>The LaserPointer attached to the opposite hand to the TabletMenu, used to interact with the TabletMenu, and select SceneObjects.</description>
        ///</item>
        ///</list>
        ///</summary>
        public void Setup()
        {
            TearDown();

            // Find annotation objects:
            //if (FindObjectsOfType<UI.ConditionMenuController>().Length != 0)
            //{
            //    foreach (UI.ConditionMenuController obj in FindObjectsOfType<UI.ConditionMenuController>())
            //    {
            //        _annotationObjects.Add(obj.gameObject);
            //    }
            //}
            if (FindObjectsOfType<UI.ConditionMenu>().Length != 0)
            {
                foreach (UI.ConditionMenu obj in FindObjectsOfType<UI.ConditionMenu>())
                {
                    _annotationObjects.Add(obj.gameObject);
                }
            }
            if (GameObject.FindGameObjectsWithTag("Annotation").Length != 0)
            {
                foreach (GameObject obj in GameObject.FindGameObjectsWithTag("Annotation"))
                {
                    if (!_annotationObjects.Contains(obj))
                    {
                        _annotationObjects.Add(obj);
                    }
                }
            }

            // Find SceneObjects:
            foreach (SceneObject obj in FindObjectsOfType<SceneObject>())
            {
                _sceneObjects.Add(obj);
            }

            if (_debug) Debug.Log("GameManager: Compiled gameobject lists."
                                            + " \tAnnontation: " + _annotationObjects.Count
                                            + " \tAITraining: " + _trainingObjects.Count
                                            + " \tSceneObjects: " + _sceneObjects.Count);

            _actor = FindObjectOfType<droid.Runtime.Prototyping.Actors.Actor>()?.gameObject;

            if (_actor != null)
                _trainingObjects.Add(_actor);

            // removed as it breaks loading
            //StopTraining(); // TODO: Look into this as BioIK creates errors when disabling panda arm.


            // SteamVR setup:
            if (usingVR)
            {
                // Finding player:
                Valve.VR.InteractionSystem.Player[] players = FindObjectsOfType<Valve.VR.InteractionSystem.Player>();
                if (players.Length == 0)
                {
                    Debug.LogError("GameManager's usingVR property is set to true, but there is no SteamVR player in the scene.");
                    noVR();
                }
                else if (players.Length >= 1)
                {
                    if (players.Length > 1)
                    {
                        Debug.LogError("There is more than one SteamVR player in the scene. Using the first one.");
                    }
                    player = players[0].transform;
                }
            }
            if (usingVR)        // If usings VR was not disabled by GameManager because of missing player.
            {
                // Finding hands:
                Valve.VR.InteractionSystem.Hand[] hands = FindObjectsOfType<Valve.VR.InteractionSystem.Hand>();
                if (hands.Length == 2)
                {
                    if (hands[0].handType == SteamVR_Input_Sources.LeftHand)
                    {
                        leftHand = hands[0];
                        rightHand = hands[1];
                    }
                    else if (hands[0].handType == SteamVR_Input_Sources.RightHand)
                    {
                        rightHand = hands[0];
                        leftHand = hands[1];
                    }
                }
                else
                {
                    Debug.LogWarning("Unexpected number of hands, needs 2 hands. Found " + hands.Length + " hands. UsingVR property in game manager now set to false.");
                    noVR();
                }
            }
            else
            {
                noVR();
            }

            //Tablet setup:
            if (tabletMenuHandFollow == null || tabletMenu == null)
            {
                var tabletMenuGameObject = FindObjectOfType<UI.TabletMenu>()?.gameObject;
                if (tabletMenuGameObject == null && _tabletPrefab != null)
                {
                    tabletMenuGameObject = Instantiate(_tabletPrefab);
                }
                if (tabletMenuGameObject != null)
                {
                    tabletMenuHandFollow = tabletMenuGameObject.GetComponent<UI.TransformFollow>();
                    tabletMenu = tabletMenuGameObject.GetComponent<UI.TabletMenu>();
                    Canvas = tabletMenu.GetComponentInChildren<Canvas>();
                }
                else
                {
                    Debug.LogError("Unable to establish TabletMenu reference");
                }
            }
            // Disableing the tablet menu since we don't wantit running on start if in vr.
            //if (usingVR && tabletMenuHandFollow != null)
            //{
            //   // tabletMenuHandFollow.isActive = false;
            //}

            // LaserPointer setup:
            if (laserPointerHandFollow == null || laserPointer == null) //NOTE: this may always be the case - redundant?
            {
                var laserPointerGameObject = FindObjectOfType<UI.LaserPointer>()?.gameObject;
                if (laserPointerGameObject == null && _laserPointerPrefab != null)
                {
                    laserPointerGameObject = Instantiate(_laserPointerPrefab);
                }
                if (laserPointerGameObject != null)
                {
                    laserPointerHandFollow = laserPointerGameObject.GetComponent<UI.TransformFollow>();
                    laserPointer = laserPointerGameObject.GetComponent<UI.LaserPointer>();
                }
                else
                {
                    Debug.LogError("Unable to establish LaserPointer reference");
                }
            }
            // Disableing the laserpointer since we don't wantit running on start if in vr.
            if (laserPointerHandFollow != null)
            {
                rayInputModule = laserPointerHandFollow.GetComponentInChildren<Internal.RaycastingInputModule>();
                rayInputModule.hand = rightHand;
                laserPointerHandFollow.followRightHand = true;
            }

            StartAnnotation();
            StartTraining();
        }

        private void TearDown()
        {
            //StartAnnotation();
            //StartTraining();
            _annotationObjects.Clear();
            _trainingObjects.Clear();
            _sceneObjects.Clear();

            tabletMenu              = null;
            leftHand                = null;
            rightHand               = null;
            laserPointer            = null;
            laserPointerHandFollow  = null;
            _actor                  = null;
        }

        private void noVR()
        {
            usingVR = false;
            if (player == null)
            {
                Debug.LogWarning("GameManager's using VR set to false, but no player was set manualy. Spawning mocked player.");
                player = new GameObject("MockPlayer").transform;
            }
        }

        //as this is a monobehaviour update is available
        private void Update()
        {
            if (_buttonTimeoutLeft > 0)
            {
                _buttonTimeoutLeft -= Time.deltaTime;
            }

            _termCount = IsTerminatingState();
            _goalCount = IsGoalState();
            if (_debug) Debug.Log("SceneObjects: " + sceneObjectsCount + " \tObjects in goalState: " + _goalCount);

            // Menu Tablet update:
            if (usingVR && _buttonTimeoutLeft <= 0 && (_interactUIButton.GetState(SteamVR_Input_Sources.LeftHand) || _interactUIButton.GetState(SteamVR_Input_Sources.RightHand)))
            {
                _buttonTimeoutLeft = _buttonTimeout;
                if (tabletMenuHandFollow.isActive)      // The menu is on, we want it off.
                {
                    tabletMenuHandFollow.isActive = false;
                    laserPointerHandFollow.GetComponentInChildren<Internal.RaycastingInputModule>().enabled = false;
                    laserPointerHandFollow.transform.GetChild(0).transform.GetChild(0).gameObject.SetActive(false);
                }
                else                                // The menu is off, we want it on.
                {
                    tabletMenuHandFollow.isActive = true;
                    laserPointerHandFollow.GetComponentInChildren<Internal.RaycastingInputModule>().enabled = true;
                    laserPointerHandFollow.transform.GetChild(0).transform.GetChild(0).gameObject.SetActive(true);

                    if (_interactUIButton.GetState(SteamVR_Input_Sources.LeftHand))
                    {
                        tabletMenuHandFollow.followRightHand = false;
                        laserPointerHandFollow.followRightHand = true;
                        rayInputModule.hand = rightHand;
                    }
                    if (_interactUIButton.GetState(SteamVR_Input_Sources.RightHand))
                    {
                        tabletMenuHandFollow.followRightHand = true;
                        laserPointerHandFollow.followRightHand = false;
                        rayInputModule.hand = leftHand;
                    }
                }
            }

        }

        private void StopAnnotation()
        {
            foreach (GameObject obj in _annotationObjects)
            {
                obj.SetActive(false);
            }
        }

        private void StopTraining()
        {
            foreach (GameObject obj in _trainingObjects)
            {
                obj.SetActive(false);
            }

        }

        /// <summary>
        /// Enables all the objects in the list of annotation objects.
        /// </summary>
        [ContextMenu("Start annotation")]
        public void StartAnnotation()
        {
            foreach (GameObject obj in _annotationObjects)
            {
                obj.SetActive(true);
            }
        }

        /// <summary>
        /// Leaves annotation mode, disableing all object realted to annotaion, then enables the actor.
        /// </summary>
        [ContextMenu("Start AI Training")]
        public void StartTraining()
        {
            StopAnnotation();
            if (_actor == null)
            {
                Debug.LogError("Actor is not set. Can not start training without an actor.");
            }
            foreach (GameObject obj in _trainingObjects)
            {
                obj.SetActive(true);
            }
        }

        /// <summary>
        /// Evaluates rather the scene is in a goal state.
        /// </summary>
        /// <returns>integer of how many SceneObjects are in a goal state.</returns>
        public int IsGoalState()
        {
            int count = 0;
            foreach (SceneObject obj in _sceneObjects)      // Loop through sceneObejcts, once we find at least one condition that is not completed.
            {                                               // Then a goal state is impossible.
                if (!obj.GoalConditions.Evaluate())         // Might not have conditions? GoalConditions?.Evaluate() maybe?
                {
                    count++;
                }
            }
            return count;
        }

        /// <summary>
        /// Evaluates rather the scene is in a terminating state.
        /// </summary>
        /// <returns>integer of how many SceneObjects are in a terminating state.</returns>
        public int IsTerminatingState()
        {
            int count = 0;
            foreach (SceneObject obj in _sceneObjects)      // Loop through sceneObejcts, once we find at least one terminating condition, we are in a terminating state.
            {
                if (obj.TerminatingConditions.ConditionsInGroup != 0 && obj.TerminatingConditions.Evaluate())
                {
                    count++;
                    if (_debug) Debug.Log("OBJECT IN TERMINATING STATE: " + obj.name);
                }
            }
            return count;                                // No object had a condition that was in a terminating state.
        }

        /// <summary>
        /// Enables or disables all the ConditionIndicators in the scene, these are the sprites hovering above SceneObjects that have one or more conditions.
        /// </summary>
        /// <param name="indicatorsEnabled">The state to set all condition indicators to, true enables them, false disables them.</param>
        public void SetConditionIndicators(bool indicatorsEnabled)
        {
            foreach (SceneObject obj in _sceneObjects)
            {
                UI.ConditionIndicator indicator = obj.GetComponent<UI.ConditionIndicator>();
                if (indicator != null)
                {
                    indicator.visible = indicatorsEnabled;
                }
            }
        }
        /// <summary>
        /// Reloads the currently active scene. Ensuring that there is only one player after the load. 
        /// </summary>
        [ContextMenu("Reload Scene")]
        public void ReloadScene()
        {
            if (_debug) Debug.Log("Reloading scene.");
            SceneManager.LoadScene(SceneManager.GetActiveScene().name);
            Destroy(((Valve.VR.InteractionSystem.Player)FindObjectOfType(typeof(Valve.VR.InteractionSystem.Player))).gameObject);
        }

        /// <summary>
        /// Removes all conditions from all SceneObjects in the scene.
        /// </summary>
        public void RemoveAllConditions()
        {
            if (_debug) Debug.Log("Removing all conditions from all objects in scene.");
            foreach (SceneObject obj in _sceneObjects)
            {
                obj.RemoveAllConditions();
            }
            // Tablet.emptyLists().
        }

    } // GameManger class.
}   // Playground Namespace.
