﻿using UnityEngine;

/// <summary>
/// Controlls the distance between two configurable joints moving linearly on the Z axis. 
/// <para name="MaxDistance">float stating the maximum distance a finger can move form the center. 0.35f is default for the panda hand from the FrankaEmikaPanda robot arm.</para>
/// </summary>
public class PandaHand : MonoBehaviour
{
    [SerializeField] private float _maxDistance = 0.35f;
    private ConfigurableJoint _fingerJointLeft;
    private ConfigurableJoint _fingerJointRight;
    private Transform _fingerBottomColliderLeft;
    private Transform _fingerBottomColliderRight;

    // Start is called before the first frame update
    private void Awake()
    {
        _fingerJointLeft = transform.Find("Fingers").Find("FingerLeft").GetComponent<ConfigurableJoint>();
        _fingerJointRight = transform.Find("Fingers").Find("FingerRight").GetComponent<ConfigurableJoint>();

        _fingerBottomColliderLeft = transform.Find("FingerColliders").Find("Left");
        _fingerBottomColliderRight = transform.Find("FingerColliders").Find("Right");

        if (_fingerBottomColliderLeft == null || _fingerBottomColliderRight == null ||
            _fingerJointLeft == null          || _fingerJointRight == null)
        {
            Debug.LogError("Panda hand is missing referenced to its children.");
        }
        gameObject.SetActive(false);
    }

    private void Update()
    {
        _fingerBottomColliderLeft.localPosition = new Vector3(  _fingerBottomColliderLeft.localPosition.x, 
                                                                _fingerBottomColliderLeft.localPosition.y, 
                                                                _fingerJointLeft.transform.localPosition.z);

        _fingerBottomColliderRight.localPosition = new Vector3( _fingerBottomColliderRight.localPosition.x, 
                                                                _fingerBottomColliderRight.localPosition.y, 
                                                                _fingerJointRight.transform.localPosition.z);
    }

    // The distance from the center to fingers should move to. 0.35 is max. As defined in their configurableJoints.
    /// <summary>
    /// Changes the spaceing of the fingers on the PandaHand. Typically called from a Steam Interactalbe object that is equipped in a hand and take some analog input.
    /// </summary>
    /// <param name="percentOpen">float between 0 and 1 controlling how open the fingers are. 0 is closed. 1 is completely open.</param>
    public void SetHandOpen(float percentOpen)
    {
        _fingerJointLeft.targetPosition = new Vector3(0, 0, percentOpen * _maxDistance);
        _fingerJointRight.targetPosition = new Vector3(0, 0, percentOpen * _maxDistance);
    }
    
}
