﻿using UnityEngine;
using Valve.VR;
using Valve.VR.InteractionSystem;

/// <summary>
/// Toggles between two panda hand GameObjects. A grab point, and the actual hand. The grab point is only a grabbable model that is equipped using SteamVR Interactable. It takes analog input from the controller, which is used to open and close the fingers.
/// The other panda hand follows the grab point using a rigidbody and slerping its movements to act similarly to a robot hand that would be controlled by motors.
/// <para name="Use Velocity">bool controlling whether the hand moves by setting position or by changing its velovity.</para>
/// <para name="Pos Alignment Speed">Speed multiplier. The speed the hand will move with towards the grabpoint which is on the VR controller.</para>
/// <para name="Max Speed">Top speed. The max speed the hand will be clamped to.</para>
/// <para name="Ang AlignmentSpeed">Angular speed multiplier. How quickly it will move towards it's desired rotation.</para>
/// </summary>
public class PandaHandGrabPoint : MonoBehaviour
{
    public SteamVR_Action_Single _actionSteering = SteamVR_Input.GetAction<SteamVR_Action_Single>("pandahand", "FingerPinch");
    private Rigidbody _pandaHandRigid;           // The rigidbody of the real hand with fingers that grabs stuff.
    private PandaHand _pandaHand;                // The script that opens and closes the fingers. 
    private GameObject _dummyHand;               // The dummy panda hand model that acts as the point the real hand tries to move to.
    private Interactable _interactable;
    [SerializeField]
    private bool _useVelocity = true;           // Use rigidbody velocity to move the hand.   
    [SerializeField]
    private float _posAlignmentSpeed = 5;
    [SerializeField]
    private float _maxSpeed = 0.5f;                 // The speed the hand will be clamped to.
    [SerializeField]
    private float _angAlignmentSpeed = 0.2f;                 // The speed the hand will rotate with towards the grabpoint which is on the VR controller.

    Vector2 _steer = Vector2.zero;
    Rigidbody _grabRigid;

    // Start is called before the first frame update
    private void Start()
    {
        _grabRigid = GetComponent<Rigidbody>();
        _pandaHandRigid = transform.parent.Find("PandaHand").GetComponent<Rigidbody>();
        _pandaHand = transform.parent.Find("PandaHand").GetComponent<PandaHand>();
        _interactable = GetComponent<Interactable>();
        _dummyHand = transform.Find("Hand").gameObject;
    }

    // Update is called once per frame
    private void Update()
    {
        if (_interactable.attachedToHand)        // Genereates error. Not set to an instance of an object. TODO
        {
            _dummyHand.SetActive(false);
            _pandaHandRigid.gameObject.SetActive(true);

           _pandaHandRigid.MoveRotation(Quaternion.Slerp(_pandaHandRigid.rotation, transform.rotation,
                    Mathf.Abs(Quaternion.Dot(_pandaHandRigid.rotation, transform.rotation) * _angAlignmentSpeed)));
            if (_useVelocity)
            {
                _pandaHandRigid.velocity = Vector3.ClampMagnitude((_grabRigid.position - _pandaHandRigid.position) * _posAlignmentSpeed, _maxSpeed);
            }
            else
            {
                _pandaHandRigid.MovePosition(transform.position);
            }

            SteamVR_Input_Sources hand = _interactable.attachedToHand.handType;
            float trigger = _actionSteering.GetAxis(hand);
            trigger = (1-trigger) / 10f;
            _pandaHand.SetHandOpen(trigger);
            
        }
        else
        {
            _dummyHand.SetActive(true);
            _pandaHandRigid.gameObject.SetActive(false);
        }
    }
}
