using System.Linq;
using UnityEngine;

namespace Playground
{
    /// <summary>
    /// container for condition, used for serializing
    /// </summary>
    [System.Serializable]
    public class ConditionContainer
    {
        public ConditionValue value;
        public uint thisId;
        public uint contextId;
        public uint relativeId;
        public bool inNestedGroup;

        public SceneObject loadedOntoObject;

        public virtual Condition LoadIntoScene()
        {
            throw new System.NotImplementedException("Condition has not yet implemented custom container"); 
        }
        /// <summary>
        /// Initializes a condition
        /// </summary>
        /// <typeparam name="TCondition">Condition type to be initialized</typeparam>
        /// <returns>condition initialized</returns>
        public TCondition Initialize<TCondition>() where TCondition : Condition
        {
            loadedOntoObject = GameManager.Instance.SceneObjects.Where(sO => sO.Id == contextId).First();
            var newCondition = loadedOntoObject.gameObject.AddComponent<TCondition>();
            newCondition.Initialize(loadedOntoObject, thisId);

            if (!inNestedGroup)
            {
                if (value == ConditionValue.goal)
                    loadedOntoObject.AddGoalCondition(newCondition);
                else
                    loadedOntoObject.AddTerminatingCondition(newCondition);
            }

            return newCondition;
        }

        /// <summary>
        /// Initializes a condition TRelative
        /// </summary>
        /// <typeparam name="TCondition">Condition type to be initialized</typeparam>
        /// <typeparam name="TRelative">Relative type</typeparam>
        /// <returns>condition initialized</returns>
        public TCondition Initialize<TCondition, TRelative>() where TCondition : Condition where TRelative : Component
        {
            var newCondition = Initialize<TCondition>();

            var relativeCondition = newCondition as Condition<TRelative>;
            if(relativeCondition != null)
            {
                if (relativeId != 0)
                {
                    var relative = GameManager.Instance.SceneObjects.Where(sO => sO.Id == relativeId).First()?.GetComponent<TRelative>();
                    if(relative != null)
                        relativeCondition.SetRelative(relative);
                    else
                        Debug.LogWarning("Loaded relative condition with invalid scene object id", relativeCondition);
                }
                else
                {
                    Debug.LogWarning("Loaded relative condition with invalid relative id", relativeCondition);
                }
            }

            return newCondition;
        }

        /// <summary>
        /// Used to copy a base container value into a new of derived type
        /// </summary>
        public T CopyBasicValues<T>() where T : ConditionContainer, new()
        {
            T returnObject = new T();
            returnObject.value         = value;
            returnObject.thisId        = thisId;
            returnObject.contextId     = contextId;
            returnObject.relativeId    = relativeId;
            returnObject.inNestedGroup = inNestedGroup;
            return returnObject;
        }
    }



    /// <summary>
    /// Used as an intermediate state for touch condition while loading
    /// </summary>
    [System.Serializable]
    public class TouchConditionContainer : ConditionContainer
    {
        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<TouchCondition, Collider>();
            return newCondition;
        }
    }

    /// <summary>
    /// Used as an intermediate state for position condition while loading
    /// </summary>
    [System.Serializable]
    public class PositionConditionContainer : ConditionContainer
    {
        public VolumeBase volumeBase;
        public PositionCondition.VolumeMode volumeMode;

        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<PositionCondition, Transform>();
            newCondition._volume = volumeBase;
            newCondition.volumeMode = volumeMode;
            return newCondition;
        }
    }

    /// <summary>
    /// Used as an intermediate state for rotation condition while loading
    /// </summary>
    [System.Serializable]
    public class RotationConditionContainer : ConditionContainer
    {
        public Vector3 TargetEulerRoation, AllowedEulerDeviation;

        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<RotationCondition, Transform>();
            newCondition.TargetEulerRoation = TargetEulerRoation;
            newCondition.AllowedEulerDeviation = AllowedEulerDeviation;
            return newCondition;
        }
    }

    /// <summary>
    /// Used as an intermediate state for velocity condition while loading
    /// </summary>
    [System.Serializable]
    public class VelocityConditionContainer : ConditionContainer
    {
        public float MaxVelocity, MinVelocity;

        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<VelocityCondition, Transform>();
            newCondition.MaxVelocity = MaxVelocity;
            newCondition.MinVelocity = MinVelocity;
            return newCondition;
        }
    }

    /// <summary>
    /// Used as an intermediate state for group condition while loading
    /// </summary>
    [System.Serializable]
    public class ConditionGroupContainer : ConditionContainer
    {
        // TODO: we probably don't need this
        public uint[] conditionsIds;

        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<ConditionGroup_OR>();
            return newCondition;
        }
    }

    /// <summary>
    /// Used as an intermediate state for or group condition while loading
    /// </summary>
    [System.Serializable]
    public class OrGroupConditionContainer : ConditionGroupContainer
    {
        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<ConditionGroup_OR>();
            return newCondition;
        }
    }

    /// <summary>
    /// Used as an intermediate state for and group condition while loading
    /// </summary>
    [System.Serializable]
    public class AndGroupConditionContainer : ConditionGroupContainer
    {
        public override Condition LoadIntoScene()
        {
            var newCondition = Initialize<ConditionGroup_AND>();
            return newCondition;
        }
    }
}
