﻿using UnityEngine.SceneManagement;
using UnityEngine;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System;
using System.Collections.Generic;
using System.Runtime.Serialization;
using UnityEditor;

namespace Playground
{
    public enum ConditionValue { goal = 1, terminating = -1 }
}


/* TODO: 
 *      - Open for custom surrogates
 *      - Make save path something that is created on load/save instead of class scope data (i.e: rework CreateFolderPath())
*/
namespace Playground.SaveSystem
{
    /// <summary>
    /// Class to handle file I/O. 
    /// Used by the save menu to serialize byte data
    /// </summary>
    ///<a href = "https://unity3d.com/learn/tutorials/topics/scripting/introduction-saving-and-loading" >class is based on the binary sub-title</a>
    // This is a singleton, and must not be initalized on load
    //[InitializeOnLoad]
    public class ConditionDataManager : MonoBehaviour
    {
        private struct SurrogateSelectorData
        {
            public ISerializationSurrogate selector;
            public Type replacedType;
        }

        /// <summary>
        /// The next id to be given to a condition
        /// </summary>
        public uint UnusedId { private set; get; }
        public void ClearIDs()
        {
            UnusedId = 1;
        }

        private List<SurrogateSelectorData> surrogateSelectors = new List<SurrogateSelectorData>();

        static ConditionDataManager _instance;
       
        /// <summary>
        /// The persistent manager state 
        /// </summary>
        public static ConditionDataManager Instance
        {
            get
            {
                if (!_instance)
                    _instance = FindObjectOfType<ConditionDataManager>();
                if (!_instance)
                    _instance = new GameObject("ConditionDataManager", typeof(ConditionDataManager)).GetComponent<ConditionDataManager>();

                return _instance;
            }
        }

        private const string _folderName = "PlaygroundScenes";
        public string FolderName => _folderName;
        private const string fileExtension = ".psbd"; // PlaygroundSceneBinaryData
        private string folderPath;
        private Scene _activeScene;

        [SerializeField] private string _selectedProfileName = "";
        public string SelectedProfileName
        {
            get { return _selectedProfileName; }
            set { _selectedProfileName = value; }
        }

        [SerializeField] private bool _loadProfileOnLoad;
        public bool LoadProfileOnLoad
        {
            get { return _loadProfileOnLoad; }
            set { _loadProfileOnLoad = value; }
        }

        private static void PlayModeStateChanged(PlayModeStateChange state)
        {
            switch (state)
            {
                case PlayModeStateChange.ExitingPlayMode:
                    Instance.SaveCurrentScene();
                    break;

                //case PlayModeStateChange.EnteredEditMode:
                default:
                    Instance.LoadCurrentScene();
                    break;
            }      
        }

        public void AddSurrogate(Type typeToReplace, ISerializationSurrogate selector)
        {
            var surrogate = new SurrogateSelectorData
            {
                selector = selector,
                replacedType = typeToReplace,
            };
            surrogateSelectors.Add(surrogate);
        }

        public void OnSceneLoaded(Scene scene, LoadSceneMode mode)
        {
            // TODO: Check a bool or something to see if it should actually load
            // OR: Some refactor must be done in SavesWindow or datamanager

            _activeScene = SceneManager.GetActiveScene();

            try
            {
                GameManager.Instance.Setup(); // Runs setup to ensure all data is correct
            }
            catch (Exception e)
            {
                Debug.LogError("GameManager threw exception:" + e.Message);
            }
            if (_loadProfileOnLoad && _selectedProfileName != "")
            {
                LoadScene(_selectedProfileName);
                _loadProfileOnLoad = false;
            }
        }


        #region LoadSaveRegion
        private void CreateFolderPath()
        {
#if UNITY_EDITOR
            _activeScene = SceneManager.GetActiveScene();
#endif
            folderPath = Path.Combine(Application.persistentDataPath, _folderName);
            folderPath = folderPath.Replace('/', '\\');
            if (!Directory.Exists(folderPath))
                Directory.CreateDirectory(folderPath);

            folderPath = Path.Combine(folderPath, _activeScene.name);
            if (!Directory.Exists(folderPath))
                Directory.CreateDirectory(folderPath);
        }

        [ContextMenu("Save current scene")] // TODO: Recover feature
        public void SaveCurrentScene()
        {
            var activeScene = SceneManager.GetActiveScene();

            SaveScene(activeScene.name + " " + DateTime.Now.ToString("yyyy-MM-dd"));
        }

        /// <summary>
        /// Used to serialize every condition in the scene 
        /// </summary>
        /// <param name="profileName">
        /// Profile to be loaded. 
        /// User can create multiple profile per scene
        /// </param>
        public void SaveScene(string profileName)
        {
            CreateFolderPath();
            string dataPath = Path.Combine(folderPath, profileName + fileExtension);

            // Create a SurrogateSelector.
            var surrogateSelector = new SurrogateSelector();

            //AddSurrogate(typeof(Vector3), new Vector3Surrogate());
            surrogateSelector.AddSurrogate(typeof(Vector2), new StreamingContext(StreamingContextStates.All), new Vector2Surrogate());
            surrogateSelector.AddSurrogate(typeof(Vector3), new StreamingContext(StreamingContextStates.All), new Vector3Surrogate());
            surrogateSelector.AddSurrogate(typeof(Matrix4x4), new StreamingContext(StreamingContextStates.All), new Matrix4x4Surrogate());
            surrogateSelector.AddSurrogate(typeof(Transform), new StreamingContext(StreamingContextStates.All), new TransformSurrogate());
            surrogateSelector.AddSurrogate(typeof(Quaternion), new StreamingContext(StreamingContextStates.All), new QuaternionSurrogate());
            surrogateSelector.AddSurrogate(typeof(SceneObject), new StreamingContext(StreamingContextStates.All), new SceneObjectSurrogate());

            var context = new StreamingContext(StreamingContextStates.All);
            foreach (var surrogate in surrogateSelectors)
            {
                surrogateSelector.AddSurrogate(surrogate.replacedType, context, surrogate.selector);
            }

            BinaryFormatter binaryFormatter = new BinaryFormatter
            {
                // Associate the SurrogateSelector with the BinaryFormatter.
                SurrogateSelector = surrogateSelector
            };
              
            // Delete old scene data
            File.Delete(dataPath);
            
            using (FileStream fileStream = File.Open(dataPath, FileMode.Create))
            {
                foreach (var sceneObject in GameManager.Instance.SceneObjects)
                {
                    SaveCondition(sceneObject.GoalConditions, ConditionValue.goal, binaryFormatter, fileStream);
                    SaveCondition(sceneObject.TerminatingConditions, ConditionValue.terminating, binaryFormatter, fileStream);
                }
            }
        }

        private void SaveCondition(ConditionGroup conditionGroup, ConditionValue conditionValue,
            BinaryFormatter binaryFormatter, FileStream fileStream, bool inNestedGroup = false)
        {
            foreach (var condition in conditionGroup)
            {
                // if we are serializing a group -
                var group = condition as ConditionGroup;
                if (group != null)
                {
                    // - we save the member conditions first
                    SaveCondition(group, conditionValue, binaryFormatter, fileStream, true);
                }

                var data = condition.GetDataContainer(conditionValue, inNestedGroup);
                try
                {
                    binaryFormatter.Serialize(fileStream, data);
                }
                catch (SerializationException e)
                {
                    Debug.LogError("failed to serialize specified data with exception message: " + e.Message);
                }
            }
        }

        [ContextMenu("Load current scene")]
        public void LoadCurrentScene()
        {
            var activeScene = SceneManager.GetActiveScene();
            LoadScene(activeScene.name);
        }

        public void LoadScene(string profileName)
        {
            if(profileName == "")
            {
                return;
            }

            CreateFolderPath();

            UnusedId = (uint)GameManager.Instance.SceneObjects.Count; 

            string dataPath = Path.Combine(folderPath, profileName + fileExtension);
            long dataLength = new FileInfo(dataPath).Length;
            if (dataLength <= 0)
            {
                Debug.LogWarningFormat("attempted to load empty file in scene: {0}", profileName);
                return;
            }
            // Create a SurrogateSelector.
            var surrogateSelector = new SurrogateSelector();

            // TODO: automate adding surrogates | StreamingContextStates.File
            surrogateSelector.AddSurrogate(typeof(Vector2), new StreamingContext(StreamingContextStates.All), new Vector2Surrogate());
            surrogateSelector.AddSurrogate(typeof(Vector3), new StreamingContext(StreamingContextStates.All), new Vector3Surrogate());
            surrogateSelector.AddSurrogate(typeof(Matrix4x4), new StreamingContext(StreamingContextStates.All), new Matrix4x4Surrogate());
            surrogateSelector.AddSurrogate(typeof(Transform), new StreamingContext(StreamingContextStates.All), new TransformSurrogate());
            surrogateSelector.AddSurrogate(typeof(Quaternion), new StreamingContext(StreamingContextStates.All), new QuaternionSurrogate());
            surrogateSelector.AddSurrogate(typeof(SceneObject), new StreamingContext(StreamingContextStates.All), new SceneObjectSurrogate());

            var context = new StreamingContext(StreamingContextStates.All);
            foreach (var surrogate in surrogateSelectors)
            {
                surrogateSelector.AddSurrogate(surrogate.replacedType, context, surrogate.selector);
            }

            BinaryFormatter binaryFormatter = new BinaryFormatter
            {
                // Associate the SurrogateSelector with the BinaryFormatter.
                SurrogateSelector = surrogateSelector
            };

            using (FileStream fileStream = File.Open(dataPath, FileMode.Open))
            {
                ConditionContainer newCondition = new ConditionContainer();
                var nonloadedConditions = new List<Condition>();
                while (fileStream.Position < dataLength)
                {
                    try
                    {
                        newCondition = (ConditionContainer)binaryFormatter.Deserialize(fileStream);
                    }
                    catch (Exception e)
                    {
                        Debug.LogWarningFormat("loading encountered an exception with message: {0}", e.Message);

                    }
                    var loadedCondition = newCondition.LoadIntoScene();

                    // if condition is part of a nested condition group
                    if (newCondition.inNestedGroup)
                        nonloadedConditions.Add(loadedCondition); // we don't convert it to a condition yet

                    // if we have loaded a group, we need to find all condition data for conditions that are members of this group and add them
                    var loadedGroup = loadedCondition as ConditionGroup;
                    if (loadedGroup != null)
                    {
                        var groupData = newCondition as ConditionGroupContainer;
                        foreach (var id in groupData?.conditionsIds ?? System.Linq.Enumerable.Empty<uint>())
                        {
                            var memberCondition = nonloadedConditions.Find(cond => cond.Id == id);
                            if (memberCondition != null)
                                loadedGroup.AddCondition(memberCondition);
                            else
                                Debug.LogWarning("Incorrect loading of condition group"); // something has become corrupt or error in loading
                        }                                                                 // @Discuss: throw exception instead?
                    }

                    // todo: validation of loading 
                    if(UnusedId < newCondition.thisId)
                    {
                        UnusedId = newCondition.thisId + 1;
                    }
                }
            }
        }
        #endregion
        #region UnityFunctions
        private void Awake()
        {
            if (Instance != this)
                Destroy(gameObject);

            DontDestroyOnLoad(gameObject);

            UnusedId = 1;

            //DontDestroyOnLoad(gameObject);
            SceneManager.sceneLoaded -= OnSceneLoaded;
            SceneManager.sceneLoaded += OnSceneLoaded;
        }

        public void DeleteProfile(string profileName)
        {
            CreateFolderPath();
            string dataPath = Path.Combine(folderPath, profileName + fileExtension);

            File.Delete(dataPath);
        }
        
        // TODO: This should be hash maybe?
        public uint GetUniqueId()
        {
            return UnusedId++;
        }

        #endregion

    }
}

