using System.Collections.Generic;

namespace Playground.UI
{
    public interface IConditionEditor
    {
        Condition GetCondition();

        //void BuildUI_2D();
        //void BuildUI_3D();

        void Initialize( Condition condition );

        //void DestroyUI_2D();
        //void DestroyUI_3D();
        void DestroyEditor();

        IReadOnlyList<Widget2D> Widget2Ds { get; }
        IReadOnlyList<WidgetVR> Widget3Ds { get; }
    }
    
    /// <summary>
    /// The editor defines how the condition parameters can be modified
    /// by constructing interactable <see cref="Widget"/>s for the user
    /// </summary>
    /// <typeparam name="TCondition">The Condition type The editor is targeted at</typeparam>
    /// <example>
    /// To notify Playground about the existance of this editor, 
    /// to allow usage of it you will need to mark the editor with
    /// <see cref="ConditionEditorAttribute"/> and specify the <see cref="System.Type"/>
    /// The target condition. This should allways be the same as <typeparamref name="TCondition"/>.
    /// <code>
    /// [ConditionEditor(typeof(SomeCondition))]
    /// public class SomeConditionEditor : ConditionEditor&lt;SomeCondition&gt; { }
    /// </code>
    /// The reason for the Generic type specifier <typeparamref name="TCondition"/> is
    /// really only to give the editor access to the values from the Condition directly,
    /// as opposed to having to cast from the base <see cref="Playground.Condition"/>.
    /// </example>
    [UnityEngine.HelpURL("https://justworksltd.gitlab.io/playground-docs/class_playground_1_1_u_i_1_1_condition_editor.html")]
    public abstract class ConditionEditor<TCondition> : Widget, IConditionEditor where TCondition : Condition
    {
        private List<Widget2D> children2D = new List<Widget2D>();
        private List<WidgetVR> children3D = new List<WidgetVR>();

        /// <summary>Gets the widget2Ds.</summary>
        /// <value>A readable list of 2D widgets the editor is owner of</value>
        public IReadOnlyList<Widget2D> Widget2Ds => children2D;

        /// <summary>Gets the widget3Ds.</summary>
        /// <value>A readable list of 3D widgets the editor is owner of</value>
        public IReadOnlyList<WidgetVR> Widget3Ds => children3D;

        /// <summary>
        /// Instantiate an instance of a 2D Widget from the <see cref="WidgetCollection"/>
        /// </summary>        
        /// <example> 
        /// This sample shows how a typical widget creation happens.
        /// <code>
        /// [ConditionEditor(typeof(SomeCondition))]
        /// public class SomeConditionEditor  : ConditionEditor&lt;SomeCondition&gt;
        /// {
        ///     ValueHandle&lt;float&gt; floatHandle;
        ///     SomeFloatWidget2D someFloatWidget; 
        ///     public override void BuildUI_2D()
        ///     {
        ///         CreateWidget2D(
        ///             out someFloatWidget,
        ///             out floatHandle,
        ///             Condition.range
        ///         );
        ///     }
        ///     
        ///     //...
        ///     
        ///     public override void OnPushChanges()
        ///     {
        ///         //push the changes up to the condition
        ///         Condition.range = floatHandle.Value;
        ///     }
        /// }
        /// </code>
        /// </example>
        /// <typeparam name="TWidget">The type of widget to be created</typeparam>
        /// <typeparam name="TValue">The value type the widget operates on (throuh the <paramref name="valueHandle"/>)</typeparam>
        /// <param name="widget">the widget to be created</param>
        /// <param name="valueHandle">a handle to the underlying data the widget is modifying</param>
        /// <param name="initialValue">for persistency, this should be the value to be edited that the condition holds</param>
        /// <param name="setOwner">set the editor to be owner of the widget, in most, if not all cases, should be <see langword="true"/></param>
        /// <param name="setup">additional setup action to be ran immediately after creation</param>
        /// <exception cref="KeyNotFoundException">Thrown if the widget typename is not found in the <see cref="WidgetCollection"/></exception>
        protected void CreateWidget2D<TWidget, TValue>(
            out TWidget widget,
            out ValueHandle<TValue> valueHandle,
            TValue initialValue,
            bool setOwner = true,
            System.Action<TWidget> setup = default
            ) where TWidget : Widget2D<TValue>
        {
            CreateWidget2D(out widget, out valueHandle, typeof(TWidget).Name, initialValue, setup);
            if (setOwner)
                children2D.Add(widget);
        }

        private static void CreateWidget2D<TWidget, TValue>( 
            out TWidget widget, 
            out ValueHandle<TValue> valueHandle, 
            string widgetName,
            TValue initialValue,
            System.Action<TWidget> setup
            ) where TWidget : Widget2D<TValue>
        {
            try
            {
                widget = Instantiate((TWidget)WidgetCollection.Instance[widgetName]); //May throw cast Exception
                widget.FetchHandle(out valueHandle);
                widget.Value = initialValue;
                if (setup != null)
                    setup.Invoke(widget);
                return;
            }
            catch (System.Exception e) { UnityEngine.Debug.LogException(e); }
            widget = null;
            valueHandle = null;
        }

        /// <summary>
        /// Instantiate an instance of a 3D Widget from the <see cref="WidgetCollection"/>
        /// </summary>        
        /// <example> 
        /// This sample shows how a typical widget creation happens.
        /// <code>
        /// [ConditionEditor(typeof(SomeCondition))]
        /// public class SomeConditionEditor : ConditionEditor&lt;SomeCondition&gt;
        /// {
        ///     ValueHandle&lt;float&gt; floatHandle;
        ///     SomeFloatWidget3D someFloatWidget; 
        ///     public override void BuildUI_3D()
        ///     {
        ///         CreateWidget3D(
        ///             out someFloatWidget,
        ///             out floatHandle,
        ///             Condition.range
        ///         );
        ///     }
        ///     
        ///     //...
        ///     
        ///     public override void OnPushChanges()
        ///     {
        ///         //push the changes up to the condition
        ///         Condition.range = floatHandle.Value;
        ///     }
        /// }
        /// </code>
        /// </example>
        /// <typeparam name="TWidget">The type of widget to be created</typeparam>
        /// <typeparam name="TValue">The value type the widget operates on (throuh the <paramref name="valueHandle"/>)</typeparam>
        /// <param name="widget">the widget to be created</param>
        /// <param name="valueHandle">a handle to the underlying data the widget is modifying</param>
        /// <param name="initialValue">for persistency, this should be the value to be edited that the condition holds</param>
        /// <param name="setOwner">set the editor to be owner of the widget, in most, if not all cases, should be <see langword="true"/></param>
        /// <param name="setup">additional setup action to be ran immediately after creation</param>
        /// <exception cref="KeyNotFoundException">Thrown if the widget typename is not found in the <see cref="WidgetCollection"/></exception>
        protected void CreateWidget3D<TWidget, TValue>( 
            out TWidget widget, 
            out ValueHandle<TValue> valueHandle,
            TValue initialValue,
            bool setOwner = true, 
            System.Action<TWidget> setup = default 
            ) where TWidget : WidgetVR<TValue>
        {
            CreateWidget3D(out widget, out valueHandle, typeof(TWidget).Name, initialValue, setup);
            if (setOwner)
                children3D.Add(widget);
        }
        private static void CreateWidget3D<TWidget, TValue>( 
            out TWidget widget, 
            out ValueHandle<TValue> valueHandle, 
            string widgetName,
            TValue initialValue,
            System.Action<TWidget> setup
            ) where TWidget : WidgetVR<TValue>
        {
            try
            {
                widget = Instantiate((TWidget)WidgetCollection.Instance[widgetName]); //May throw cast Exception
                widget.FetchHandle(out valueHandle);
                widget.Value = initialValue;
                if (setup != null)
                    setup.Invoke(widget);
                return;
            }
            catch (System.Exception e) { UnityEngine.Debug.LogException(e); }
            widget = null;
            valueHandle = null;
        }
        /// <summary> The Condition that is beeing interfaced by this editor. </summary>
        protected new TCondition Condition {
            get => (TCondition)base.Condition;
            set => base.Condition = value;
        }
        /// <summary> Build all <see cref="Widget2D"/>s needed to modify the condition.</summary>
        protected abstract void BuildUI_2D();
        /// <summary> Build all <see cref="WidgetVR"/>s needed to modify the condition.</summary>
        protected abstract void BuildUI_3D();

        /// <summary> 
        /// Destroy all the <see cref="Widget2D"/>s this editor owns
        /// needs to be overridden if you have widgets that don't have an <see cref="Widget.Owner"/> set
        /// <para>(By default <see cref="Widget"/> Creation sets owner to the editor creating them.)</para>
        /// </summary>
        protected virtual void DestroyUI_2D()
        {
            foreach (var child in children2D)
                Destroy(child.gameObject);
        }
        /// <summary> 
        /// Destroy all the <see cref="Widget2D"/>s this editor owns
        /// needs to be overridden if you have widgets that don't have an <see cref="Widget.Owner"/> set
        /// <para>(By default <see cref="Widget"/> Creation sets owner to the editor creating them.)</para>
        /// </summary>
        protected virtual void DestroyUI_3D()
        {
            foreach (var child in children3D)
                Destroy(child.gameObject);
        }
        /// <summary>
        /// Causes creation of interactable widgets and binding of condition parameters.
        /// </summary>
        /// <param name="condition">The <see cref="Playground.Condition"/> to interface with</param>
        public void Initialize(Condition condition)
        {
            SetCondition(condition);
            BuildUI_2D();
            BuildUI_3D();

            foreach (var child in children3D)
                child.SetOwner(this);
            foreach (var child in children2D)
                child.SetOwner(this);

            OnPushChanges();
        }

        /// <summary>Gets the condition.</summary>
        /// <returns>the condition</returns>
        public Condition GetCondition() => Condition;
        private void SetCondition( Condition condition ) => base.Condition = condition;

        /// <summary>
        /// Destroy the editor and all <see cref="Widget"/>s owned by the editor
        /// </summary>
        //TODO: Maybe change over to the Destroy() -> OnDestroy pattern?
        //      To prevent dangling widgets
        public void DestroyEditor()
        {
            DestroyUI_2D();
            DestroyUI_3D();
            Destroy(gameObject);
        }
        
        public override void SetInitialState() { }
    }
}