﻿using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

using Valve.VR.InteractionSystem;
namespace Playground.UI {

    [System.Serializable]
    public struct LayoutData
    {
        /// <summary>
        /// A reference to the child beeing layed out
        /// </summary>
        public SubMenu menuRef;
        /// <summary>
        /// The position of this child when placed inside the parent submenu
        /// </summary>
        public Vector3 internalPosition;
        public Vector2 internalSizeDelta;
        //TODO: Add internalRotation?
    }


    [RequireComponent(typeof(Interactable))]
    [System.Serializable]  
    public class SubMenu : MonoBehaviour
    {
#pragma warning disable 0649
        [SerializeField] private string _name;
        public string Name
        {
            get => _name;
            set => _name = value;
        }

        [SerializeField] private SubMenu _owner;
        private SubMenu _root;

        [SerializeField] private LayoutData[] _childLayouts;
        [SerializeField] private Widget[] _widgets; //NOTE: may not need to be serialized
        [SerializeField] private Transform _grabbable; //i.e. the transform of the collider
#pragma warning restore 0649

        public SubMenu Root
        {
            get
            {
                if (!_root)
                {
                    for (SubMenu parent = _owner; parent; parent = parent._owner) {
                        _root = parent;
                    }
                }
                Debug.Assert(_root != null, $"{name}: Root was null when trying to access it!");
                return _root;
            }
        }

        [Range(50, 1000), Tooltip("Drag Threshold in pixelspace: how far you need to drag the element in order to detach it")]
        [SerializeField]
        private int _detachThreshold = 100;

        [SerializeField] RectTransform container;

        bool attachedToHand;

        [EnumFlags, Tooltip("The flags used to attach this object to the hand.")]
        public Hand.AttachmentFlags attachmentFlags = Hand.AttachmentFlags.ParentToHand | Hand.AttachmentFlags.DetachFromOtherHand;

        Vector3 startGrabPosition;

        RectTransform RectTransform => (RectTransform)transform;

        Vector3 OwnerSpace( Vector3 worldPosition ) {
            return _owner?.transform.InverseTransformPoint(worldPosition) ?? worldPosition;
        }

#if UNITY_EDITOR
        [ContextMenu("Store current Child Layout")]
        private void SaveLayout()
        {
            //TODO: Remove LayoutGroup-related things
            void RecurseTransform(ref List<SubMenu> children, Transform innerTransform)
            {
                for (int i = 0; i < innerTransform.childCount; i++)
                {
                    var child = innerTransform.GetChild(i);
                    var childMenu = innerTransform.GetChild(i).GetComponent<SubMenu>();
                    if (childMenu)
                        children.Add(childMenu);
                    else
                    {
                        var layoutGroup = child.GetComponent<LayoutGroup>();
                        if(layoutGroup)
                            RecurseTransform(ref children, layoutGroup.transform);
                    }
                }
            } 
            List<SubMenu> childList = new List<SubMenu>();
            RecurseTransform(ref childList, transform);


            _childLayouts = new LayoutData[childList.Count];
            for (int i = 0; i < _childLayouts.Length; i++)
            {
                childList[i]._owner = this;

                _childLayouts[i].menuRef = childList[i];
                _childLayouts[i].internalPosition = _childLayouts[i].menuRef.transform.localPosition;
                _childLayouts[i].internalSizeDelta = _childLayouts[i].menuRef.RectTransform.sizeDelta;
                _childLayouts[i].menuRef.SaveLayout();
            }
            UnityEditor.EditorUtility.SetDirty(this);
        }

        [ContextMenu("Test Revert")] 
        public void TestRevert()
        {
            RevertLayoutToDefault();
        }

        private void Reset()
        {
            if (!gameObject.GetComponentInParent<SubMenu>() && _root != null)
            {
                Debug.LogError("invalid menu structure: parent does not contain a submenu");
            }
        }
#endif
        /// <summary>
        /// Handles layout position data
        /// </summary>
        /// <param name="placableWidget">widget to be place</param>
        public void InsertAndLayoutWidget(Widget2D placableWidget)
        {
            placableWidget.transform.SetParent(transform, true);
            placableWidget.transform.localScale = Vector3.one;
            //TODO: MORE
        }

        private void Start()
        {
            _grabbable.Find("MenuLabel").GetComponent<Text>().text = _name;
        }

        protected void RevertLayoutToDefault()
        {
            foreach(var childLayout in _childLayouts)
            {
                childLayout.menuRef.RevertLayoutToDefault();

                if (childLayout.menuRef._owner == this)
                    continue;

                childLayout.menuRef._owner = this;
                var oldContainer = childLayout.menuRef.transform.parent;
                childLayout.menuRef.transform.parent = transform;
                Destroy(oldContainer.gameObject);
                childLayout.menuRef.transform.localPosition = childLayout.internalPosition;
                childLayout.menuRef.RectTransform.sizeDelta = childLayout.internalSizeDelta;
                childLayout.menuRef.transform.localRotation = Quaternion.identity;
            }
        }

        protected void DetachFromOwner()
        {
            //do some stuff, then:
            container = (RectTransform)(new GameObject($"{name}_container", typeof(Canvas), typeof(CanvasScaler)).transform);

            container.transform.localPosition = _owner.transform.localPosition; //TODO: move the container about to avoid floating point errors
            container.transform.localRotation = _owner.transform.localRotation;
            container.transform.localScale = Root.container.localScale;
            

            /*
             Container holds information about how to recover layout of detached parts
             Dictionary might be used for this

            OnDetachMenu should copy the submenu, reset the old one's position and disable it.
             */

            var ownerCanvasScaler = Root.container.GetComponent<CanvasScaler>();
            var canvasScaler = container.GetComponent<CanvasScaler>();

            Vector3 preUnparentPosition = transform.position;
            _owner = null;
            //container.sizeDelta = RectTransform.sizeDelta; Look at making the sizes nicer
            transform.parent = container;

            canvasScaler.dynamicPixelsPerUnit = ownerCanvasScaler.dynamicPixelsPerUnit;
            canvasScaler.referencePixelsPerUnit = ownerCanvasScaler.referencePixelsPerUnit;


            OnDetachFromOwner();
        }

        public virtual void OnDetachFromOwner()
        {

        }

        protected void SnapBack()
        {
            //snap back to the grabbing position relative to the attachment point of this menu
            transform.localPosition = startGrabPosition - _grabbable.transform.localPosition;
        }

        public virtual void OnSnapBack()
        {

        }

        #region STEAM_VR
        //-------------------------------------------------
        // Attach To Hand
        //-------------------------------------------------
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when the hand tries to make this attach</para>
        /// </summary>
        /// <param name="hand">the attaching hand</param>
        protected void OnAttachedToHand( Hand hand )
        {
            startGrabPosition = OwnerSpace(_grabbable.position);
            if (_owner)
            {
                Debug.DrawLine(_owner.transform.TransformPoint(startGrabPosition) - _owner.transform.forward, _owner.transform.TransformPoint(startGrabPosition) + _owner.transform.forward, Color.red, 20.0f);
            }
            attachedToHand = true;
            hand.HoverLock(null);
        }

        //-------------------------------------------------
        // Hand Hover Begin
        //-------------------------------------------------
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when a hand starts hovering over this</para>
        /// </summary>
        /// <param name="hand">the hovering hand</param>
        protected void OnHandHoverBegin( Hand hand )
        {
            if (!attachedToHand)
            {
                GrabTypes bestGrabType = hand.GetBestGrabbingType();
                if (bestGrabType != GrabTypes.None)
                {
                    hand.AttachObject(gameObject, bestGrabType, attachmentFlags);
                }
            }
        }

        //-------------------------------------------------
        // Hand Hover End
        //-------------------------------------------------
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when a hand stops hovering over this</para>
        /// </summary>
        /// <param name="hand">the hovering hand</param>
        protected void OnHandHoverEnd( Hand hand ) => hand.HideGrabHint();

        //-------------------------------------------------
        // Hand Hover Update
        //-------------------------------------------------
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens during updates while hovering over this</para>
        /// </summary>
        /// <param name="hand">the hovering hand</param>
        protected void HandHoverUpdate( Hand hand )
        {
            GrabTypes startingGrabType = hand.GetGrabStarting();
            if (startingGrabType != GrabTypes.None)
            {
                hand.AttachObject(gameObject, startingGrabType, attachmentFlags);
                hand.HideGrabHint();
            }
        }


        //-------------------------------------------------
        // Detach From Hand
        //-------------------------------------------------
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when this is detached</para>
        /// </summary>
        /// <param name="hand">the detaching hand</param>
        protected void OnDetachedFromHand( Hand hand )
        {
            if (_owner)
            {
                Debug.Log(Vector3.Distance(OwnerSpace(_grabbable.position), startGrabPosition));

                if (Vector3.Distance(OwnerSpace(_grabbable.position), startGrabPosition) > _detachThreshold)
                    DetachFromOwner();
                else
                    SnapBack();
            }


            attachedToHand = false;
            hand.HoverUnlock(null);
        }

        //-------------------------------------------------
        // Attached Update
        //-------------------------------------------------
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens in the update loop when this is attached</para>
        /// </summary>
        /// <param name="hand">the attached hand</param>
        protected void HandAttachedUpdate( Hand hand )
        {
            if (hand.IsGrabEnding(gameObject))
            {
                hand.DetachObject(gameObject);
            }
        }

        //-------------------------------------------------
        // Hand Focus
        //-------------------------------------------------
        protected void OnHandFocusAcquired( Hand hand ) => gameObject.SetActive(true);
        protected void OnHandFocusLost( Hand hand ) => gameObject.SetActive(false);
        #endregion
    }
}