﻿using UnityEngine;

namespace Playground.UI
{

    /// <summary>
    /// Follows its target 
    /// <para name="Pos Alignment Speed">Speed multiplier. The speed the hand will move with towards the grabpoint which is on the VR controller.</para>
    /// <para name="Pos Alignment Deadzone">A margin of position that the TrasformFollow will accept as having reached it's goal. Allowing minute movemonets to be ignored.</para>
    /// <para name="Ang Alignment Speed">Angular speed multiplier. How quickly it will move towards it's desired rotation.</para>
    /// <para name="Ang Alignment Deadzone">A margin of rotation that the TrasformFollow will accept as having reached it's goal. Allowing minute movemonets to be ignored.</para>
    /// <para name="OverrideUseNonVRTarget">When checked the Transform follow will ignore the state of the VR hands, and use the non-VR attachment point instead.</para>
    /// <para name="Non VR Attachment Target">A transfrom the TrasformFollow will follow instead of the VR hands if OverrideUseNonVRTarget is true, or if the VR hands can not be found.</para>
    /// </summary>
    public class TransformFollow : MonoBehaviour
    {
#pragma warning disable 0649
        public bool followRightHand = false;
        [SerializeField] private float _posAlignmentSpeed = 15;
        [SerializeField] private float _posAlignmentDeadzone = 0.02f;
        [SerializeField] private float _angAlignmentSpeed = 0.3f;
        [SerializeField] private float _angAlignmentDeadzone = 6;
        [SerializeField] private bool _overrideUseNonVRTarget = false;
        [SerializeField] private Transform _nonVRAttachmentTarget;      // Needs to be set if there is no VR player in the scene. This replaces the hand.
#pragma warning restore 0649

        private bool active;
        /// <summary>
        /// Property that disables all child objects. It will still however, follow it's target. 
        /// </summary>
        public bool isActive
        {
            get
            {   return active;  }
            set
            {
                active = value;
                for (int i = 0; i < transform.childCount; i++)
                {
                    transform.GetChild(i).gameObject.SetActive(active);
                }
            }
        }
 
        private Vector3 _initialPosition;
        private Quaternion _initialRotation;
        private Transform _leftHand;
        private Transform _rightHand;
        Transform target;

        private void Start()
        {
            _initialPosition = transform.position;
            _initialRotation = transform.rotation;

            if (!_overrideUseNonVRTarget && GameManager.Instance.usingVR)
            {
                _leftHand = GameManager.Instance.leftHand.transform;
                _rightHand = GameManager.Instance.rightHand.transform;
            }
            else
            {
                if (_nonVRAttachmentTarget != null)
                {
                    target = _nonVRAttachmentTarget;
                }
                else if (GameManager.Instance.player != null)
                {
                    target = GameManager.Instance.player;
                }
                else
                {
                    Debug.LogWarning("TransformFollow does not have an attachmentTarget.");
                }
            }
        }

        private void Update()
        {
            if (!_overrideUseNonVRTarget && GameManager.Instance.usingVR)
            {
                if (followRightHand && _rightHand != null)
                {
                    target = _rightHand;
                }
                else if (_leftHand != null)
                {
                    target = _leftHand;
                }
                else
                {
                    Debug.LogWarning("Game manager's usingVR property set to true, but does not have reference to a hand.");
                    return;
                }
            }

            if ((target.position - transform.position).magnitude > _posAlignmentDeadzone)
            {
                Vector3 vec = target.position + target.TransformVector(_initialPosition) - transform.position;
                transform.position += (vec - vec.normalized * _posAlignmentDeadzone) * _posAlignmentSpeed * Time.deltaTime;
            }

            //Rotation, slerping if outside deadzone:
            float t = Mathf.Abs(Quaternion.Dot(transform.rotation, target.rotation * _initialRotation));
            transform.rotation = Quaternion.Slerp(transform.rotation, Quaternion.RotateTowards((target.rotation * _initialRotation), transform.rotation, _angAlignmentDeadzone), t * _angAlignmentSpeed);
        }

        [ContextMenu("ToggleActive")]
        private void ToggleActive()
        {   isActive = !isActive;   }

    }
}


