﻿using UnityEngine;
using UnityEngine.EventSystems;
using Valve.VR.InteractionSystem;
using Valve.VR;
using System;


namespace Playground.Internal
{
    //https://bitbucket.org/Unity-Technologies/ui/src/0651862509331da4e85f519de88c99d0529493a5/UnityEngine.UI/EventSystem/InputModules/StandaloneInputModule.cs?at=2018.3%2Fstaging&fileviewer=file-view-default
    //https://github.com/wacki/Unity-VRInputModule/blob/master/Assets/VRInputModule/Scripts/LaserPointerInputModule.cs

    [RequireComponent(typeof(Camera))]
    public class RaycastingInputModule : BaseInputModule
    {
        public Hand hand;

        private float m_PrevActionTime;
        private Vector2 m_LastMoveVector;
        //private int m_ConsecutiveMoveCount;

        private Vector2 _lastPointerPosition;
        private Vector2 _pointerPosition;


#if UNITY_EDITOR
        public bool debug;
        public bool fallbackMock => !GameManager.Instance.usingVR;
        private Transform _mockHandTransform => GameManager.Instance.player;//=> _hand.transform;
        private Transform handTransform => fallbackMock ? _mockHandTransform : hand?.transform ?? _mockHandTransform;
# else
        private Transform handTransform => _hand.transform;
#endif

        public SteamVR_Input_Sources InputSource => hand?.handType ?? SteamVR_Input_Sources.Any;
#pragma warning disable 0649
        [SerializeField] private LayerMask interactionLayer;
        [SerializeField] private SteamVR_Action_Boolean _clickAction;
#pragma warning restore 0649
        //[SerializeField] private SteamVR_Action_Vector2 _scrollAction;
#if UNITY_EDITOR
        [Space]
#endif
        [SerializeField] private float _laserReach = 100f;
        private Camera _eventProcessingCamera;

        [NonSerialized] private GameObject currentEnter;
        [NonSerialized] private GameObject currentPressed;
        [NonSerialized] private GameObject currentDragging;

        RaycastHit[] sceneObjectHits = new RaycastHit[1];


        protected override void Awake()
        {
            base.Awake();
            _eventProcessingCamera = GetComponent<Camera>();
            // We don't really care about using the camera for rendering,
            // so don't do graphics related things
            _eventProcessingCamera.orthographic = true;
            _eventProcessingCamera.orthographicSize = 0.01f;
            _eventProcessingCamera.clearFlags = CameraClearFlags.Nothing;
            _eventProcessingCamera.enabled = false;
            _eventProcessingCamera.fieldOfView = 5;
            _eventProcessingCamera.nearClipPlane = 0.01f;
            _eventProcessingCamera.farClipPlane = _laserReach;
            //_eventProccessingCamera.cullingMask = CULL NON UI?

            inputOverride = GetComponent<SteamVRInput>();
        }

        protected override void Start()
        {
            base.Start();
            GameManager.Instance.Canvas.worldCamera = _eventProcessingCamera;
        }

        protected bool SendUpdateEventToSelectedObject()
        {
            if (eventSystem.currentSelectedGameObject == null)
                return false;

            var data = GetBaseEventData();
            ExecuteEvents.Execute(eventSystem.currentSelectedGameObject, data, ExecuteEvents.updateSelectedHandler);
            return data.used;
        }

        public override void Process()
        {
            bool usedEvent = SendUpdateEventToSelectedObject();
            ProcessVRPointerEvents();
        }

        protected void ProcessVRPointerEvents()
        {
            UpdateEventProcessingCamera();
            var eventData = GetVRPointerEventData();
            ProcessMove(eventData);
            ProcessClick(eventData);
        }

        private void UpdateEventProcessingCamera()
        {
            _eventProcessingCamera.transform.SetPositionAndRotation(
                handTransform.position,
                handTransform.rotation
            );
        }

        private VRPointerEventData GetVRPointerEventData()
        {
            VRPointerEventData eventData = new VRPointerEventData(eventSystem){ inputSource = InputSource };
            //TODO: cache
            eventData.Reset();
            eventData.position = _eventProcessingCamera.pixelRect.center;

            eventData.pointerPress = currentPressed;
            eventData.pointerDrag = currentDragging;
            eventData.pointerEnter = currentEnter;

            eventSystem.RaycastAll(eventData, m_RaycastResultCache); // fill the eventdata by casting a ray
            eventData.pointerCurrentRaycast = FindFirstRaycast(m_RaycastResultCache);
            //NOTE: it might be worth to do this every frame BEFORE the raycast instead,
            //      in case we want the other results for some logic
            m_RaycastResultCache.Clear();

            _pointerPosition = eventData.pointerCurrentRaycast.worldPosition;
            eventData.delta = _pointerPosition - _lastPointerPosition;
            _lastPointerPosition = _pointerPosition;

            //Debug.Log($"INIT DELTA: {eventData.delta}");

            return eventData;
        }


        public void ClearSelected()
        {
            if (eventSystem.currentSelectedGameObject) {
                eventSystem.SetSelectedGameObject(null);
            }
        }

        private static bool ShouldStartDrag( Vector2 pressPos, Vector2 currentPos, float threshold, bool useDragThreshold )
        {
            if (!useDragThreshold)
                return true;
            return (pressPos - currentPos).sqrMagnitude >= threshold * threshold;
        }

        private void ProcessMove( VRPointerEventData e )
        {
            var targetGO = e.pointerCurrentRaycast.gameObject;
            //TODO: Message the sender (controller) of the event
            //Pass OnEnter/OnExit calls down to the UI (for hover effects?)
            // IPointerEnterHandler and IPointerExitHandler
            if (ExecuteEvents.CanHandleEvent<IPointerEnterHandler>(targetGO))
            {
                if (currentEnter != targetGO)
                {
                    //if (currentEnter != null)
                    {
                        currentEnter = null;
                        //ExecuteEvents.Execute(currentEnter, e, ExecuteEvents.pointerExitHandler);
                    }

                    if (targetGO != null)
                    {
                        currentEnter = targetGO;
                        //ExecuteEvents.Execute(currentEnter, e, ExecuteEvents.pointerEnterHandler);
                    }
                }
            }
            HandlePointerExitAndEnter(e, targetGO);
            currentEnter = e.pointerEnter;
        }

        private void ProcessClick( VRPointerEventData e )
        {
            var currentOverGo = e.pointerCurrentRaycast.gameObject;
#if UNITY_EDITOR
            if ((fallbackMock && Input.GetMouseButtonDown(0)) || (!fallbackMock && _clickAction.GetStateDown(InputSource)))
#else
            if (_clickAction.GetStateDown(InputSource))
#endif
            {
                e.pressPosition = e.position;
                e.pointerPressRaycast = e.pointerCurrentRaycast;
                e.pointerPress = null;

                //if hitting a physics-interactable
                if(Physics.RaycastNonAlloc(
                    transform.position,
                    transform.forward,
                    sceneObjectHits,
                    _laserReach,
                    interactionLayer.value,
                    QueryTriggerInteraction.Ignore) > 0)
                {
                    var sceneObjectGO = sceneObjectHits[0].transform.gameObject;
                    HandlePointerExitAndEnter(e, sceneObjectGO);
                    currentEnter = e.pointerEnter;
                    if (ExecuteEvents.CanHandleEvent<IPointerClickHandler>(sceneObjectGO))
                    {
#if PLAYGROUND_DEBUG
                        Debug.Log($"POINTER CLICK on {sceneObjectGO}");
#endif
                        ExecuteEvents.Execute(sceneObjectGO, e, ExecuteEvents.pointerClickHandler);
                    }
                }
                else if (e.pointerCurrentRaycast.gameObject != null)
                {
                    GameObject newPressed = ExecuteEvents.ExecuteHierarchy(currentOverGo, e, ExecuteEvents.pointerDownHandler); //Execute a pointerDownEvent on the target
                    //TODO: Message the sender (controller) of the event
                    if (newPressed == null) //the target was not a handler of IPointerDownEvent, though it might still be one for IClickHandler
                    {
                        newPressed = ExecuteEvents.ExecuteHierarchy(currentOverGo, e, ExecuteEvents.pointerClickHandler);
#if PLAYGROUND_DEBUG
                        if (newPressed != null) Debug.Log($"POINTER CLICK on {newPressed}");
#endif
                    }
                    else
                    {
#if PLAYGROUND_DEBUG
                        Debug.Log($"POINTER DOWN on {newPressed}");
                        if(ExecuteEvents.Execute(newPressed, e, ExecuteEvents.pointerClickHandler))
                        {
                            Debug.Log($"POINTER CLICK on {newPressed}");
                        }
#else
                        ExecuteEvents.Execute(newPressed, e, ExecuteEvents.pointerClickHandler);
#endif
                        //TODO: Message the sender (controller) of the event
                    }
                    if (newPressed != null)
                    {
                        e.pointerPress = newPressed;
                        currentPressed = newPressed;

                        e.pointerDrag = newPressed;
                        currentDragging = newPressed;

                        ClearSelected();

                        if (ExecuteEvents.CanHandleEvent<ISelectHandler>(newPressed))
                        {
#if PLAYGROUND_DEBUG
                            Debug.Log($"SELECT on {newPressed}");
#endif
                            eventSystem.SetSelectedGameObject(newPressed);
                        }
#if PLAYGROUND_DEBUG
                        if(ExecuteEvents.Execute(newPressed, e, ExecuteEvents.initializePotentialDrag))
                        {
                            Debug.Log($"INIT POTENTIAL DRAG on {newPressed}");
                        }
#else
                        ExecuteEvents.Execute(newPressed, e, ExecuteEvents.initializePotentialDrag);
#endif
                    }
                }
            }
            ProcessDrag(e);

#if UNITY_EDITOR
            if ((fallbackMock && Input.GetMouseButtonUp(0)) || (!fallbackMock && _clickAction.GetStateUp(InputSource)))
#else
            if (_clickAction.GetStateUp(InputSource))
#endif
            {

                if (e.pointerPress)
                {
#if PLAYGROUND_DEBUG
                    if(ExecuteEvents.Execute(e.pointerPress, e, ExecuteEvents.pointerUpHandler))
                    {
                        Debug.Log($"POINTER UP on {e.pointerPress}");
                    }
#else
                    ExecuteEvents.Execute(e.pointerPress, e, ExecuteEvents.pointerUpHandler);
#endif
                    var pointerUpHandler = ExecuteEvents.GetEventHandler<IPointerClickHandler>(currentOverGo);
                    if (e.dragging && e.pointerDrag != null && pointerUpHandler != e.pointerPress && e.eligibleForClick)
                    {
#if PLAYGROUND_DEBUG
                        if(ExecuteEvents.ExecuteHierarchy(currentOverGo, e, ExecuteEvents.dropHandler) != null)
                        {
                            Debug.Log($"DROP on {currentOverGo}");
                        }
#else
                        ExecuteEvents.ExecuteHierarchy(currentOverGo, e, ExecuteEvents.dropHandler);
#endif
                    }

                    e.eligibleForClick = false;
                    e.pointerPress = null;
                    currentPressed = null;
                    e.rawPointerPress = null;

                    if (e.pointerDrag != null && e.dragging)
                    {
#if PLAYGROUND_DEBUG
                        if(ExecuteEvents.Execute(e.pointerDrag, e, ExecuteEvents.endDragHandler))
                        {
                            Debug.Log($"END DRAG on {e.pointerDrag}");
                        }
#else
                        ExecuteEvents.Execute(e.pointerDrag, e, ExecuteEvents.endDragHandler);
#endif
                    }

                    e.dragging = false;
                    e.pointerDrag = null;
                    currentDragging = null;

                    // redo pointer enter / exit to refresh state
                    // so that if we moused over somethign that ignored it before
                    // due to having pressed on something else
                    // it now gets it.
                    if (currentOverGo != e.pointerEnter)
                    {
                        HandlePointerExitAndEnter(e, null);
                        HandlePointerExitAndEnter(e, currentOverGo);
                    }

                }
            }
        }
        private void ProcessDrag( VRPointerEventData pointerEvent )
        {

            bool moving = pointerEvent.IsPointerMoving();
            //Debug.Log($"Pointer has {((pointerEvent.pointerDrag != null) ? "a" : "no ")} dragging target");
            //Debug.Log($"Pointer should {(ShouldStartDrag(pointerEvent.pressPosition, pointerEvent.position, eventSystem.pixelDragThreshold, pointerEvent.useDragThreshold) ? "" : "not ")} drag!");

            if (moving && pointerEvent.pointerDrag != null
                && !pointerEvent.dragging
                && ShouldStartDrag(pointerEvent.pressPosition, pointerEvent.position, eventSystem.pixelDragThreshold, pointerEvent.useDragThreshold))
            {

                pointerEvent.dragging = ExecuteEvents.Execute(pointerEvent.pointerDrag, pointerEvent, ExecuteEvents.beginDragHandler);
#if PLAYGROUND_DEBUG
                if(pointerEvent.dragging) Debug.Log($"BEGIN DRAG on {pointerEvent.pointerDrag}");
#endif
            }

            // Drag notification
            if (pointerEvent.dragging && moving && pointerEvent.pointerDrag != null)
            {
                // Before doing drag we should cancel any pointer down state
                // And clear selection!
                if (pointerEvent.pointerPress != pointerEvent.pointerDrag)
                {

#if PLAYGROUND_DEBUG
                    if(ExecuteEvents.Execute(pointerEvent.pointerPress, pointerEvent, ExecuteEvents.pointerUpHandler))
                    {
                        Debug.Log($"POINTER UP on {pointerEvent.pointerPress}");
                    }

#else
                    ExecuteEvents.Execute(pointerEvent.pointerPress, pointerEvent, ExecuteEvents.pointerUpHandler);
#endif
                    pointerEvent.eligibleForClick = false;
                    pointerEvent.pointerPress = null;
                    currentPressed = null;
                    pointerEvent.rawPointerPress = null;
                }
#if PLAYGROUND_DEBUG
                if(ExecuteEvents.Execute(pointerEvent.pointerDrag, pointerEvent, ExecuteEvents.dragHandler))
                {
                    Debug.Log($"POINTER DRAG on {pointerEvent.pointerDrag}");
                }
#else
                ExecuteEvents.Execute(pointerEvent.pointerDrag, pointerEvent, ExecuteEvents.dragHandler);
#endif
            }
        }
    }
}
