﻿using UnityEngine;
using Playground.UI.Util;
namespace Playground.UI
{

    public class CuboidWidget3D : WidgetVR<CuboidVolume>
    {
#pragma warning disable 0649
        [SerializeField] private float _scaleHandleOffset = 0.1f;
        [SerializeField] private Material _activeMat, _inactiveMat;
        [SerializeField] private MeshRenderer _meshRenderer;
        [SerializeField] private Transform _cubeTransform;
        [SerializeField] private ScaleAxisWidget3D _front, _back, _left, _right, _bottom, _top;
#pragma warning restore 0649
        public enum Face
        {
            Front   = 0,
            Back    = 1,
            Left    = 2,
            Right   = 3,
            Bottom  = 4,
            Top     = 5
        }

        /// <summary>
        /// Allow outside to control the handles if need be
        /// </summary>
        /// <param name="face">the scale axis index to change handle for</param>
        /// <param name="handle">the new handle the axis widget should use</param>
        public void SetScaleAxisHandle( Face face, in ValueHandle<float> handle )
        {
            switch (face)
            {
                case Face.Front:
                    _front.SetValueHandle(handle);
                break;
                case Face.Back:
                    _back.SetValueHandle(handle);
                break;
                case Face.Left:
                    _left.SetValueHandle(handle);
                break;
                case Face.Right:
                    _right.SetValueHandle(handle);
                break;
                case Face.Top:
                    _top.SetValueHandle(handle);
                break;
                case Face.Bottom:
                    _bottom.SetValueHandle(handle);
                break;
            }
        }

#if UNITY_EDITOR
        protected override void Reset()
        {
            base.Reset();
            _cubeTransform = (transform.Find("Mesh") ?? transform.GetChild(0));
            _meshRenderer = _cubeTransform.GetComponent<MeshRenderer>();
            
            _front = (transform.Find("Front") ?? transform.GetChild(1)).GetComponent<ScaleAxisWidget3D>();
            _front._owner = this;
            _back = (transform.Find("Back") ?? transform.GetChild(2)).GetComponent<ScaleAxisWidget3D>();
            _back._owner = this;
            _left = (transform.Find("Left") ?? transform.GetChild(3)).GetComponent<ScaleAxisWidget3D>();
            _left._owner = this;
            _right = (transform.Find("Right") ?? transform.GetChild(4)).GetComponent<ScaleAxisWidget3D>();
            _right._owner = this;
            _bottom = (transform.Find("Bottom") ?? transform.GetChild(5)).GetComponent<ScaleAxisWidget3D>();
            _bottom._owner = this;
            _top = (transform.Find("Top") ?? transform.GetChild(6)).GetComponent<ScaleAxisWidget3D>();
            _top._owner = this;

            if (_inactiveMat)
                _meshRenderer.sharedMaterial = _inactiveMat;
        }
#endif

        float[] lastValues = new float[6];

        public override void SetInitialState()
        {
            WidgetUtil.PlaceOnTopOfRelative(this);
            _cubeTransform.localScale = Condition.Context.transform.localScale;
        }

        protected override void OnConditionContextSet()
        {
            if(Value == null)
            {
#if PLAYGROUND_DEBUG
                Debug.Log("WIDGET: Context condition uninitialized! Setting it up to default");
#endif
                SetInitialState();
            }
            else
            {
#if PLAYGROUND_DEBUG
                Debug.Log("WIDGET: Restoring Cuboid Widget");
#endif
                
                transform.localRotation = Value.cuboidSpace.rotation;
                transform.localPosition = Value.cuboidSpace.GetPosition();
                _cubeTransform.localScale = Value.max - Value.min;
            }
            
            _front.transform.localPosition  = _front.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * 0.5f, _front.LocalAxis) + _scaleHandleOffset);
            _back.transform.localPosition   = _back.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * -0.5f, _back.LocalAxis) - _scaleHandleOffset);
            _left.transform.localPosition   = _left.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * -0.5f, _left.LocalAxis) - _scaleHandleOffset);
            _right.transform.localPosition  = _right.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * 0.5f, _right.LocalAxis) + _scaleHandleOffset);
            _bottom.transform.localPosition = _bottom.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * -0.5f, _bottom.LocalAxis) - _scaleHandleOffset);
            _top.transform.localPosition    = _top.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * 0.5f, _top.LocalAxis) + _scaleHandleOffset);
            
            Vector3 frontBackScale = new Vector3(_cubeTransform.localScale.x, _cubeTransform.localScale.y, 1);
            Vector3 leftRightScale = new Vector3(1, _cubeTransform.localScale.y, _cubeTransform.localScale.z);
            Vector3 bottomTopScale = new Vector3(_cubeTransform.localScale.x, 1, _cubeTransform.localScale.z);

            _front.transform.localScale = frontBackScale;
            _back.transform.localScale  = frontBackScale;

            _left.transform.localScale      = leftRightScale;
            _right.transform.localScale     = leftRightScale;

            _bottom.transform.localScale    = bottomTopScale;
            _top.transform.localScale       = bottomTopScale;
        }
        public override void OnStartUpdate()
        {
            _meshRenderer.sharedMaterial = _activeMat;
            if (Attached)
            {
                //TODO: investigate if this will this ruin the rotation fix of the scaleAxes...
                _front.gameObject.SetActive(false);
                _back.gameObject.SetActive(false);
                _left.gameObject.SetActive(false);
                _right.gameObject.SetActive(false);
                _bottom.gameObject.SetActive(false);
                _top.gameObject.SetActive(false);
            }
        }
        public override void OnUpdateVisuals()
        {
            float[] values = {
                _front.Value,
                _back.Value,
                _left.Value,
                _right.Value,
                _bottom.Value,
                _top.Value
            };

            float[] deltaValues = new float[6];
            for (int i = 0; i < deltaValues.Length; i++)
            {
                deltaValues[i] = values[i] - lastValues[i];
            }

            _cubeTransform.localPosition = _cubeTransform.localPosition + new Vector3(
                (deltaValues[(int)Face.Left] + deltaValues[(int)Face.Right])*0.5f,
                (deltaValues[(int)Face.Bottom] + deltaValues[(int)Face.Top])*0.5f,
                (deltaValues[(int)Face.Front] + deltaValues[(int)Face.Back])*0.5f 
            );

            _cubeTransform.localScale = _cubeTransform.localScale + new Vector3(
                (deltaValues[(int)Face.Right] - deltaValues[(int)Face.Left]),
                (deltaValues[(int)Face.Top] - deltaValues[(int)Face.Bottom]),
                (deltaValues[(int)Face.Front] - deltaValues[(int)Face.Back])
            );

            for (int i = 0; i < lastValues.Length; i++)
            {
                lastValues[i] = values[i];
            }
        }
        public override void OnStopUpdate()
        {
            transform.position = _cubeTransform.position;
            _cubeTransform.localPosition = Vector3.zero;

            _cubeTransform.localScale = new Vector3(
                Mathf.Abs(_cubeTransform.localScale.x), 
                Mathf.Abs(_cubeTransform.localScale.y), 
                Mathf.Abs(_cubeTransform.localScale.z)
            );
            
            _meshRenderer.sharedMaterial = _inactiveMat;
            
            _front.transform.localPosition  = _front.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * 0.5f, _front.LocalAxis) + _scaleHandleOffset);
            _back.transform.localPosition   = _back.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * -0.5f, _back.LocalAxis) - _scaleHandleOffset);
            _left.transform.localPosition   = _left.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * -0.5f, _left.LocalAxis) - _scaleHandleOffset);
            _right.transform.localPosition  = _right.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * 0.5f, _right.LocalAxis) + _scaleHandleOffset);
            _bottom.transform.localPosition = _bottom.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * -0.5f, _bottom.LocalAxis) - _scaleHandleOffset);
            _top.transform.localPosition    = _top.LocalAxis * (Vector3.Dot(_cubeTransform.localScale * 0.5f, _top.LocalAxis) + _scaleHandleOffset);
            
            Vector3 frontBackScale = new Vector3(_cubeTransform.localScale.x, _cubeTransform.localScale.y, 1);
            Vector3 leftRightScale = new Vector3(1, _cubeTransform.localScale.y, _cubeTransform.localScale.z);
            Vector3 bottomTopScale = new Vector3(_cubeTransform.localScale.x, 1, _cubeTransform.localScale.z);

            _front.transform.localScale = frontBackScale;
            _back.transform.localScale = frontBackScale;

            _left.transform.localScale = leftRightScale;
            _right.transform.localScale = leftRightScale;

            _bottom.transform.localScale = bottomTopScale;
            _top.transform.localScale = bottomTopScale;


            //TODO: Maybe only do this when the cuboid core handle was the active part?
            _front.gameObject.SetActive(true);
            _back.gameObject.SetActive(true);
            _left.gameObject.SetActive(true);
            _right.gameObject.SetActive(true);
            _bottom.gameObject.SetActive(true);
            _top.gameObject.SetActive(true);
        }

        public override void OnPushChanges()
        {
            Value = new CuboidVolume(
                new Vector3(
                    -(_cubeTransform.localScale.x) * 0.5f,
                    -(_cubeTransform.localScale.y) * 0.5f,
                    -(_cubeTransform.localScale.z) * 0.5f
                ),
                new Vector3(
                    (_cubeTransform.localScale.x) * 0.5f,
                    (_cubeTransform.localScale.y) * 0.5f,
                    (_cubeTransform.localScale.z) * 0.5f
                ),
                transform.worldToLocalMatrix
            );
        }

        private void OnDestroy()
        {
            Destroy(gameObject);
        }

    }
}
