﻿using Playground.UI.Util;
using UnityEngine;
namespace Playground.UI
{
    /// <summary>
    /// Sphere Widget 3D with a float radius
    /// </summary>
    [System.Serializable]
    public class CylinderWidget3D : WidgetVR<CylinderVolume>
    {
        [SerializeField] private PointWidget3D _rangePoint;
        [SerializeField] private Transform _meshTransform;
        [SerializeField] private ScaleAxisWidget3D _top, _bottom;
        [SerializeField] private float _scaleHandleOffset = 0.1f;
        float[] lastValues = new float[2];

        public enum Face { Top = 0, Bottom = 1 }

        /// <summary>
        /// Allow outside to control the handles if need be
        /// </summary>
        /// <param name="face">the scale axis index to change handle for</param>
        /// <param name="handle">the new handle the axis widget should use</param>
        public void SetScaleAxisHandle( Face face, in ValueHandle<float> handle )
        {
            switch(face)
            {
                case Face.Top:
                    _top.SetValueHandle(handle);
                break;
                case Face.Bottom:
                    _bottom.SetValueHandle(handle);
                break;
            }
        }

        /// <summary>
        /// Allow outside to control the Range point value handle
        /// </summary>
        /// <param name="handle">the new handle the widget should use</param>
        public void SetRangePointHandle( in ValueHandle<Vector3> handle ) => _rangePoint.SetValueHandle(handle);

#if UNITY_EDITOR
        /* Only compiled when in unity editor, allows us to automatically 
         * get the wanted references upon adding/reseting the component
         */
        protected override void Reset()
        {
            base.Reset();
            _meshTransform = transform.Find("Mesh") ?? transform.GetChild(0);
            _rangePoint = transform.GetChild(1).GetComponent<PointWidget3D>();
            _rangePoint._owner = this;

            _top = (transform.Find("Top") ?? transform.GetChild(2)).GetComponent<ScaleAxisWidget3D>();
            _top._owner = this;

            _bottom = (transform.Find("Bottom") ?? transform.GetChild(3)).GetComponent<ScaleAxisWidget3D>();
            _bottom._owner = this;
        }
#endif
        public override void SetInitialState()
        {
            //NOTE: would it be more sensible to place this one inside? ...
            //      I think so, as it will probably be used more for physical containers!
            WidgetUtil.PlaceInsideRelative(this);

            //TODO: Add RelativeBounds-obtainer in WidgetUtil
            var contextScale = Condition.Context.transform.localScale;
            Vector3 rangeScaler = new Vector3(contextScale.x*0.5f, 0, 0);
            _rangePoint.transform.position = transform.TransformPoint(rangeScaler);
            _meshTransform.localScale = new Vector3(rangeScaler.magnitude * 2, contextScale.y, rangeScaler.magnitude * 2);
        }

        protected override void OnConditionContextSet()
        {
            if (Value == null)
            {
#if PLAYGROUND_DEBUG
                Debug.Log("WIDGET: Context condition uninitialized! Setting it up to default");
#endif
                SetInitialState();
            }
            else
            {
#if PLAYGROUND_DEBUG
                Debug.Log("WIDGET: Restoring Cuboid Widget");
#endif
                _rangePoint.transform.position = Value.radius * transform.right;
                _rangePoint.transform.localScale = Vector3.one * (Value.radius * 0.25f + 0.125f);
                transform.position = Value.center;
                _meshTransform.localScale = new Vector3(Value.radius * 2, Value.height, Value.radius * 2);
                transform.localRotation = Value.cylinderSpace.rotation;
            }

            Vector3 rangePointLocal = transform.InverseTransformPoint(_rangePoint.transform.position);
            float radius = Vector3.Distance(Vector3.up * Vector3.Dot(rangePointLocal, Vector3.up), rangePointLocal);
            
            _top.transform.localPosition = _top.LocalAxis * (_meshTransform.localScale.y * 0.5f + _scaleHandleOffset);
            _bottom.transform.localPosition = _bottom.LocalAxis * (_meshTransform.localScale.y * -0.5f - _scaleHandleOffset);

            _bottom.transform.localScale = new Vector3(radius * 2, 1, radius * 2);
            _top.transform.localScale = new Vector3(radius * 2, 1, radius * 2);
        }

        public override void OnStartUpdate()
        {
            //TODO: do some visual stuff with materials
        }
        public override void OnUpdateVisuals()
        {
            Vector3 rangePointLocal = transform.InverseTransformPoint(_rangePoint.transform.position);
            float radius = Vector3.Distance(Vector3.up * Vector3.Dot(rangePointLocal, Vector3.up), rangePointLocal);
            
            float[] deltaValues = new float[] {
                _top.Value - lastValues[(int)Face.Top],
                _bottom.Value - lastValues[(int)Face.Bottom]
            };

            _meshTransform.localPosition = _meshTransform.localPosition + new Vector3(
                0, (deltaValues[(int)Face.Bottom] + deltaValues[(int)Face.Top]) * 0.5f, 0
            );

            _meshTransform.localScale = new Vector3(
                radius * 2,
                _meshTransform.localScale.y + (deltaValues[(int)Face.Top] - deltaValues[(int)Face.Bottom]),
                radius * 2
            );

            lastValues[(int)Face.Top] = _top.Value;
            lastValues[(int)Face.Bottom] = _bottom.Value;
            _rangePoint.transform.localScale = Vector3.one * (Value.radius * 0.25f + 0.125f);
        }

        public override void OnStopUpdate()
        {
            Vector3 rangePointLocal = transform.InverseTransformPoint(_rangePoint.transform.position);
            Vector3 clampedRangePointLocal = new Vector3(rangePointLocal.x, 0, rangePointLocal.z);
            _rangePoint.transform.position = transform.TransformPoint(clampedRangePointLocal);
            float radius =  clampedRangePointLocal.magnitude;

            transform.position = _meshTransform.position;
            _meshTransform.localPosition = Vector3.zero;

            _meshTransform.localScale = new Vector3(
                radius * 2,
                Mathf.Abs(_meshTransform.localScale.y),
                radius * 2
            );

            _top.transform.localPosition = _top.LocalAxis * (_meshTransform.localScale.y * 0.5f + _scaleHandleOffset);
            _bottom.transform.localPosition = _bottom.LocalAxis * (_meshTransform.localScale.y * -0.5f - _scaleHandleOffset);

            _bottom.transform.localScale = new Vector3(radius * 2, 1, radius * 2);
            _top.transform.localScale = new Vector3(radius * 2, 1, radius * 2);
            _rangePoint.transform.localScale = Vector3.one * (Value.radius * 0.25f + 0.125f);
        }

        public override void OnPushChanges()
        {
            float radius = transform.InverseTransformPoint(_rangePoint.transform.position).magnitude;
            Value = new CylinderVolume(
                center: transform.position,
                height: _meshTransform.localScale.y,
                radius: radius,
                cylinderSpace: transform.worldToLocalMatrix
            );
        }
    }
}