﻿using UnityEngine;

namespace Playground.UI.Util
{
    /// <summary>
    /// Utilities for placement of widgets
    /// </summary>
    public static class WidgetUtil
    {
        public static void PlaceInsideRelative(Widget widget)
        {
            if(widget.Condition.HasRelative)
            {
                Transform relative = ((Internal.IHasRelative)widget.Condition).Relative.transform;
                widget.transform.SetPositionAndRotation(relative.position, relative.rotation);
                return;
            }//Fallback to place near context
            widget.transform.position = widget.Condition.Context.transform.position + Vector3.up * widget.Condition.Context.transform.localScale.y * 1.25f;
        }

        public static void PlaceOnTopOfRelative(Widget widget)
        {
            if(widget.Condition.HasRelative)
            {
                Component relative = ((Internal.IHasRelative)widget.Condition).Relative;

                Vector3 upDirection = GetLikelyLocalDirection(relative.transform, Vector3.up);
                Vector3 localUpDirection = relative.transform.InverseTransformDirection(upDirection);

                Bounds bounds = relative.GetComponent<MeshFilter>()?.sharedMesh.bounds ?? new Bounds(relative.transform.position, Vector3.zero);
                foreach (var mf in relative.GetComponentsInChildren<MeshFilter>())
                {
                    //TODO: transform modelspace of nested meshes into local space of relative

                    //bounds.Encapsulate(relative.transform.TransformPoint(mf.sharedMesh.bounds.max));
                    //bounds.Encapsulate(relative.transform.TransformPoint(mf.sharedMesh.bounds.min));
                }
                
                widget.transform.position = relative.transform.position + relative.transform.TransformVector(localUpDirection * bounds.extents.y) + (upDirection* widget.Condition.Context.transform.localScale.y * 0.5f);
                widget.transform.rotation = relative.transform.rotation;
                return;
            } // Fallback to place near Context

            widget.transform.position = widget.Condition.Context.transform.position + Vector3.up * widget.Condition.Context.transform.localScale.y * 1.25f;
        }

        const float SQRT_VEC3_ONE = 0.57736f;
        /// <summary>
        /// Get the direction relative to a transform that is likely it's variant of the <paramref name="targetDirection"/> 
        /// based on the dot product of the transform's local axes.
        /// <para>
        /// in other words; the possible outcomes are:
        /// <list type="bullet">
        ///     <item><term> t.up       </term></item>
        ///     <item><term>-t.up       </term></item>
        ///     <item><term> t.right    </term></item>
        ///     <item><term>-t.right    </term></item>
        ///     <item><term> t.forward  </term></item>
        ///     <item><term>-t.forward  </term></item>
        /// </list>       
        /// </para>
        /// </summary>
        /// <param name="t">the transform</param>
        /// <param name="targetDirection">the target axis you want to find in localSpace of the transform</param>
        /// <returns>the vector that is most likely defining </returns>
        public static Vector3 GetLikelyLocalDirection( in Transform t, Vector3 targetDirection )
        {
            Vector3 targetDir = targetDirection.normalized;
            Vector3 direction = Vector3.up;
            float maxDot = 0;
            float dot = Vector3.Dot(t.up, targetDir);
            if (Mathf.Abs(maxDot) < Mathf.Abs(dot))
            {
                maxDot = dot;
                direction = t.up;
                //check if the transform-local axis is closer to the target direction than the other ones.
                if (Mathf.Abs(maxDot) > SQRT_VEC3_ONE)
                    return Mathf.Sign(maxDot) * direction;
            }

            dot = Vector3.Dot(t.forward, targetDir);
            if (Mathf.Abs(maxDot) < Mathf.Abs(dot))
            {
                maxDot = dot;
                direction = t.forward;
                //check if the transform-local axis is closer to the target direction than the other ones.
                if (Mathf.Abs(maxDot) > SQRT_VEC3_ONE) 
                    return Mathf.Sign(maxDot) * direction;
            }

            dot = Vector3.Dot(t.right, targetDir);
            if (Mathf.Abs(maxDot) < Mathf.Abs(dot))
            {
                maxDot = dot;
                direction = t.right;
            }
            return Mathf.Sign(maxDot) * direction;
        }
    }
}
