﻿using UnityEngine;
using Valve.VR.InteractionSystem;

namespace Playground.UI
{
    public class ValueHandle<T>
    {
        public ValueHandle( T value ) { Value = value; }
        public ValueHandle() { Value = default; }
        public T Value { get; set; }
        public static implicit operator T( ValueHandle<T> handle ) => handle.Value;
    }
    [HelpURL("https://justworksltd.gitlab.io/playground-docs/class_playground_1_1_u_i_1_1_widget.html")]
    public abstract class Widget : MonoBehaviour
    {
#if UNITY_EDITOR
        protected virtual void Reset() { }
#endif
        [SerializeField] protected internal Widget _owner;
        public Widget Owner => _owner;
        public Condition Condition { get; protected set; }

        public void SetOwner( Widget owner )
        {
            if (owner == this)
            {
                Debug.LogError($"{this} tried to set owner to self!");
                return;
            }
            _owner = owner;
            Condition = owner.Condition;
            OnConditionContextSet();
            PropogatePushChanges(); //ensure the Widget in its finalized initial state is pushed up 
        }

        /// <summary>
        /// Override to do component specific setup based on condition context.
        /// This is usually where you would ensure persistency by reconstructing
        /// the widget settings based on the value handle of the widget
        /// </summary>
        protected virtual void OnConditionContextSet() { }

        /// <summary>Sets the widget to its initial state.</summary>
        public abstract void SetInitialState();
        protected void PropagateStartUpdate()
        {
            OnStartUpdate();
            _owner?.PropagateStartUpdate();
        }
        protected void PropogateUpdateVisuals()
        {
            OnUpdateVisuals();
            _owner?.PropogateUpdateVisuals();
        }
        protected void PropogateStopUpdate()
        {
            OnStopUpdate();
            _owner?.PropogateStopUpdate();
        }
        protected void PropogatePushChanges()
        {
            OnPushChanges();
            _owner?.PropogatePushChanges();
        }

        /// <summary>
        /// Called upon when grabbed/attached
        /// </summary>
        public virtual void OnStartUpdate() { }

        /// <summary>
        /// Updates done internally
        /// </summary>
        public virtual void OnUpdateVisuals() { }

        /// <summary>
        /// Called upon when released/un-attached
        /// </summary>
        public virtual void OnStopUpdate() { }

        /// <summary>
        /// update external state so that the rest of the UI-system reccognize the changes made.
        /// </summary>
        public virtual void OnPushChanges() { }
    }

    //TODO deal with handshaking widget manipulation somehow
    //TODO maybe add SteamVR_SkeletonPoser as required too
    //TODO strip away unused bloat (steamVR-related)
    [RequireComponent(typeof(Interactable))]
    public abstract class WidgetVR : Widget
    {
#if UNITY_EDITOR
        protected override void Reset() => _interactable = GetComponent<Interactable>();
#endif
        [SerializeField] protected Interactable _interactable;
        protected bool attached;
        public bool Attached => attached;

        [EnumFlags]
        [Tooltip("The flags used to attach this object to the hand.")]
        public Hand.AttachmentFlags attachmentFlags = Hand.AttachmentFlags.ParentToHand | Hand.AttachmentFlags.DetachFromOtherHand;


        //=========================================================
        //------------------------STEAM VR-------------------------
        //=========================================================

        //-------------------------------------------------
        // Attach To Hand
        //-------------------------------------------------
        /// <summary>
        /// Default implementation of hand attachment for widgets
        /// </summary>
        /// <param name="hand">the hand to attach to</param>
        protected virtual void DoAttachToHand( Hand hand )
        {
            attached = true;
            hand.HoverLock(null);
        }
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when the hand tries to make this attach</para>
        /// </summary>
        /// <param name="hand">the attaching hand</param>
        protected void OnAttachedToHand( Hand hand )
        {
            DoAttachToHand(hand);
            PropagateStartUpdate();
        }

        //-------------------------------------------------
        // Hand Hover Begin
        //-------------------------------------------------
        /// <summary>
        /// Default implementation of hand beggining hovering for widgets
        /// </summary>
        /// <param name="hand">the hand hovering over this</param>
        protected virtual void DoHandHoverBegin( Hand hand )
        {
            if (!attached)
            {
                GrabTypes bestGrabType = hand.GetBestGrabbingType();
                if (bestGrabType != GrabTypes.None)
                {
                    hand.AttachObject(gameObject, bestGrabType, attachmentFlags);
                }
            }
        }
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when a hand starts hovering over this</para>
        /// </summary>
        /// <param name="hand">the hovering hand</param>
        protected void OnHandHoverBegin( Hand hand ) => DoHandHoverBegin(hand);

        //-------------------------------------------------
        // Hand Hover End
        //-------------------------------------------------
        /// <summary>
        /// Default implementation of hand stopping hovering for widgets
        /// </summary>
        /// <param name="hand">the hand hovering over this</param>
        protected virtual void DoHandHoverEnd( Hand hand ) => hand.HideGrabHint();
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when a hand stops hovering over this</para>
        /// </summary>
        /// <param name="hand">the hovering hand</param>
        protected void OnHandHoverEnd( Hand hand ) => DoHandHoverEnd(hand);

        //-------------------------------------------------
        // Hand Hover Update
        //-------------------------------------------------
        /// <summary>
        /// Default implementation of updates while hovering over widgets
        /// </summary>
        /// <param name="hand">the hand hovering over this</param>
        protected virtual void DoHandHoverUpdate( Hand hand )
        {
            GrabTypes startingGrabType = hand.GetGrabStarting();
            if (startingGrabType != GrabTypes.None)
            {
                hand.AttachObject(gameObject, startingGrabType, attachmentFlags);
                hand.HideGrabHint();
            }
        }
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens during updates while hovering over this</para>
        /// </summary>
        /// <param name="hand">the hovering hand</param>
        protected void HandHoverUpdate( Hand hand ) => DoHandHoverUpdate(hand);


        //-------------------------------------------------
        // Detach From Hand
        //-------------------------------------------------
        /// <summary>
        /// Default implementation of detaching widgets from hand
        /// </summary>
        /// <param name="hand">the detaching hand</param>
        protected virtual void DoDetachFromHand( Hand hand )
        {
            attached = false;
            PropogateStopUpdate();
            PropogatePushChanges();
            hand.HoverUnlock(null);
        }
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens when this is detached</para>
        /// </summary>
        /// <param name="hand">the detaching hand</param>
        protected void OnDetachedFromHand( Hand hand ) => DoDetachFromHand(hand);


        //-------------------------------------------------
        // Attached Update
        //-------------------------------------------------
        /// <summary>
        /// Default implementation of attached updates for widgets
        /// </summary>
        /// <param name="hand">the attached hand</param>
        protected virtual void DoAttachedUpdate( Hand hand )
        {
            PropogateUpdateVisuals();
            if (hand.IsGrabEnding(gameObject))
            {
                hand.DetachObject(gameObject);
            }
        }
        /// <summary>
        /// Valve's interface for interactables
        /// <para>What happens in the update loop when this is attached</para>
        /// </summary>
        /// <param name="hand">the attached hand</param>
        protected void HandAttachedUpdate( Hand hand ) => DoAttachedUpdate(hand);

        //-------------------------------------------------
        // Hand Focus
        //-------------------------------------------------
        protected void OnHandFocusAcquired( Hand hand ) => gameObject.SetActive(true);
        protected void OnHandFocusLost( Hand hand ) => gameObject.SetActive(false);
    }

    public abstract class WidgetVR<T> : WidgetVR, IValuedWidget<T>
    {
        private ValueHandle<T> valueHandle = new ValueHandle<T>();
        public void FetchHandle( out ValueHandle<T> handle ) => handle = valueHandle;
        public void SetValueHandle( ValueHandle<T> handle ) => valueHandle = handle;
        public T Value
        {
            get => valueHandle.Value;
            set => valueHandle.Value = value;
        }
    }


    public abstract class Widget2D : Widget
    {
        public RectTransform RectTransform => transform as RectTransform;
    }
    public abstract class Widget2D<T> : Widget2D, IValuedWidget<T>
    {
        private ValueHandle<T> valueHandle = new ValueHandle<T>();

        public void FetchHandle( out ValueHandle<T> handle ) => handle = valueHandle;
        public void SetValueHandle( ValueHandle<T> handle ) => valueHandle = handle;
        public T Value
        {
            get => valueHandle.Value;
            set => valueHandle.Value = value;
        }
    }
}